"use strict";

/* Copyright (c) 2012 Andy McDonald. All rights reserved. */
/* Please refer to licence.txt for licensing terms. */

// ====================================================================================
Components.utils.import("resource://foxclocks/engine.js");

// ====================================================================================
function FoxClocks()
{		
	this.zonePicker = null;
	this.watchlistTree = null;
	this.searchBox = null;
	this.watchlistDropAtIndex = null; // AFM - null => drop at end
	this.timeFormatter = new FoxClocks_TimeFormatter();
	this.inOnZonePickerSelect = false;
	this.clockUpdateTimer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);

	this.ZP_DST_INDICATOR_STRING = "*";
}
	
// ====================================================================================
FoxClocks.prototype =
{
	// ====================================================================================
	onLoad : function()
	{
		this.zonePicker = document.getElementById("fc-zonepicker");
		this.watchlistTree = document.getElementById("fc-watchlist");
		this.searchBox = document.getElementById("fc-zonepicker-searchbox");
		this.searchBox.setAttribute("class", "fc-zonepicker-searchbox-inactive");
		this.setTimeFormat();
		
		FoxClocks_Utils.getFoxClocksVersion(function(foxClocksVersion) {
			document.getElementById("fc-foxclocks-version-label").setAttribute("value", "FoxClocks" + ' ' + foxClocksVersion);
		});

		// AFM - missing locale strings
		//
		if (this.searchBox.getAttribute("fc_init_value") == "")
			this.searchBox.setAttribute("collapsed", "true");
			
		var self = this;
		FoxClocks_Engine.onStarted(function() { self.startup(); } );
	},
	
	// ====================================================================================
	startup : function()
	{
		FoxClocks_Logger.log("+FoxClocks::startup()");

		this.populateZonePicker();
		this.populateWatchlist();
		
		this.clockUpdateTimer.init(this, FoxClocks_Utils.FC_CLOCK_UPDATE_INTERVAL, Components.interfaces.nsITimer.TYPE_REPEATING_PRECISE);
		
		FoxClocks_PrefManager.addPrefObserver("foxclocks.", this);
		
		var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
		observerService.addObserver(this, "foxclocks", false);
	
		FoxClocks_Logger.log("-FoxClocks::startup()");
	},
	
	// ====================================================================================
	onUnload : function()
	{
		this.clockUpdateTimer.cancel()
			
		FoxClocks_PrefManager.removePrefObserver("foxclocks.", this);
		
		var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
		try {observerService.removeObserver(this, "foxclocks");} catch(ex){}
	},
	
	// AFM - START, EVENTS
	//
	
	// ====================================================================================
	onWatchListKeyPress : function(event)
	{
		if (event.keyCode == KeyEvent.DOM_VK_ENTER || event.keyCode == KeyEvent.DOM_VK_RETURN)
			this.onOpenLocDetailsCmd();
	},

	// ====================================================================================
	onZonePickerSelect : function()
	{
		// AFM - horrible way to disable selecting non-terminal nodes
		//
		
		// AFM - prevent recursion. this.zonePicker.view.selection.selectEventsSuppressed
		// doesn't seem to work inside an event handler - in fact causes recursion.
		// We're called back again when we clear a container's selection, so can't use
		// toggleSelection since we'll reselect the container. Yuk.
		//
		if (this.inOnZonePickerSelect)
			return;
		
		this.inOnZonePickerSelect = true;
	
		FoxClocks_Logger.log("+FoxClocks::onZonePickerSelect()");
		
		var rangeCount = this.zonePicker.view.selection.getRangeCount();
		
		for (var i=0; i < rangeCount; i++)
		{
			var start = {};
			var end = {};
			this.zonePicker.view.selection.getRangeAt(i, start, end);
			for (var c = start.value; c != -1 && c <= end.value; c++)
			{
				var treeItem = this.zonePicker.view.getItemAtIndex(c);

				if (treeItem.getAttribute("container") == "true")
					this.zonePicker.view.selection.clearRange(c, c);
			}
		}
		
		this.setZonePickerStates();
		this.inOnZonePickerSelect = false;
		
		FoxClocks_Logger.log("-FoxClocks::onZonePickerSelect()");
	},
	
	// ====================================================================================
	onOpenZoneDetailsCmd : function()
	{
		var location = this.getSelectedZonePickerLocation();
		if (!location)
			return;
					
		// AFM - ignore return value
		//
		var wathlistItem = new FoxClocks_WatchlistItem(location);
		this.openZoneInfo(wathlistItem, "ZONE");
	},
	
	// ====================================================================================
	onOpenZoneGoogleEarthCmd : function()
	{
		var location = this.getSelectedZonePickerLocation();
		
		if (location)
			FoxClocks_Utils.openGoogleEarth(location);
	},
	
	// ====================================================================================
	onImportCmd : function()
	{
		var dialogTitle = document.getElementById("fc-dialog-import-title").getAttribute("label");
		var foxClocksFilter = "*." + FoxClocks_Utils.FC_FOXCLOCKS_SETTINGS_EXTENSION;
		var filterText = document.getElementById("fc-dialog-importexport-filter-label").getAttribute("label") +
			" (" + foxClocksFilter + ")";

		var nsIFilePicker = Components.interfaces.nsIFilePicker;
		var fp = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
		fp.init(window, dialogTitle, nsIFilePicker.modeOpen);
		fp.appendFilter(filterText, foxClocksFilter);
		fp.appendFilters(nsIFilePicker.filterAll);
		
		var res = fp.show();
		
		if (res != nsIFilePicker.returnOK)
			return;
	
		var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
		
		try
		{
			var req = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
			req.open('GET', fp.fileURL.spec + '?time=' + new Date().getTime(), true);
			req.addEventListener('loadend', function() {
		
				try 
				{
					if (req.responseXML === null)
						throw 'Bad response';
	
					if (req.responseXML.documentElement.nodeName === "parsererror")
						throw req.responseXML.documentElement.firstChild.nodeValue;
							
					var promptText = document.getElementById("fc-import-confirm-label").getAttribute("label");  	
					if (promptService.confirm(window, "FoxClocks", promptText) == true)
					{
						if (FoxClocks_PrefManager.xmlToPrefs("foxclocks.", req.responseXML) == false)
							throw 'parsererror';
						
						var promptText = document.getElementById("fc-dialog-import-success").getAttribute("label");
						FoxClocks_Utils.openChromeSimpleInfo(window, "FoxClocks", promptText);
					}
				}
				catch(ex)
				{
					var promptText = document.getElementById("fc-dialog-import-failure").getAttribute("label") + ": " + ex;
					promptService.alert(window, "FoxClocks", promptText);
				}
				
			}, false);
			
			req.send(null);
		}
		catch(ex) 
		{
			FoxClocks_Logger.error("FoxClocks::onImportCmd(): " + ex);
			var promptText = document.getElementById("fc-dialog-import-failure").getAttribute("label") + ": " + ex;
			promptService.alert(window, "FoxClocks", promptText);
		}
	},

	// ====================================================================================
	onExportCmd : function()
	{
		var dialogTitle = document.getElementById("fc-dialog-export-title").getAttribute("label");
		var foxClocksFilter = "*." + FoxClocks_Utils.FC_FOXCLOCKS_SETTINGS_EXTENSION;
		var filterText = document.getElementById("fc-dialog-importexport-filter-label").getAttribute("label") +
			" (" + foxClocksFilter + ")";								

		var nsIFilePicker = Components.interfaces.nsIFilePicker;
		var fp = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
		fp.init(window, dialogTitle, nsIFilePicker.modeSave);
		fp.defaultExtension = FoxClocks_Utils.FC_FOXCLOCKS_SETTINGS_EXTENSION;
		fp.defaultString = "foxclocks." + FoxClocks_Utils.FC_FOXCLOCKS_SETTINGS_EXTENSION;
		fp.appendFilter(filterText, foxClocksFilter);
		fp.appendFilters(nsIFilePicker.filterAll);
		
		var res = fp.show();
		
		if (res != nsIFilePicker.returnOK && res != nsIFilePicker.returnReplace)
			return;
	
		FoxClocks_Utils.getFoxClocksVersion(function(foxClocksVersion) {
	
			var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
		
			try
			{
				var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].createInstance(Components.interfaces.nsIDOMSerializer);
				var xmlDoc = FoxClocks_PrefManager.prefsToXml("foxclocks.", null, "prefs", foxClocksVersion);
				var xmlString = FoxClocks_Utils.FC_XML_DEC + serializer.serializeToString(xmlDoc, "UTF-8");

				FoxClocks_Utils.writeToFile(xmlString, fp.file, function(result, success) {
					
					if (success)
					{
						var promptText = document.getElementById("fc-dialog-export-success").getAttribute("label");
						FoxClocks_Utils.openChromeSimpleInfo(window, "FoxClocks", promptText);
					}
					else
					{
						var promptText = document.getElementById("fc-dialog-export-failure").getAttribute("label") + ": " + result;
						promptService.alert(window, "FoxClocks", promptText);
					}
				});
			} 
			catch(ex) 
			{
				var promptText = document.getElementById("fc-dialog-export-failure").getAttribute("label") + ": " + ex;
				promptService.alert(window, "FoxClocks", promptText);
			}
		});
	},
		
	// ====================================================================================
	onAddCmd : function(addAtIndex, reSort)
	{
		FoxClocks_Logger.log("+FoxClocks::onAddCmd()");
												
		var rangeCount = this.zonePicker.view.selection.getRangeCount();
		var addedItems = new Array();
		
		var zonePickerLocMap = FoxClocks_ZoneManager.getZonePickerLocationMap();
		
		for (var i=0; i < rangeCount; i++)
		{
			var start = {};
			var end = {};
			this.zonePicker.view.selection.getRangeAt(i, start, end);
			for (var c = start.value; c != -1 && c <= end.value; c++)
			{
				var j = this.zonePicker.view.getItemAtIndex(c).getAttribute("value");
				if (j != null)
				{
					var watchlistItem = new FoxClocks_WatchlistItem(zonePickerLocMap[j]);
					addedItems.push(watchlistItem);
				}
			}
		}
		
		this.addToWatchlist(addedItems, addAtIndex);
		
		if (reSort != null && reSort == false)
		{
			// AFM - currently removes the sort marker even if the order hasn't changed
			//
			this.setWatchlistUnsorted();
		}
		else
		{
			this.sortWatchlist(); // AFM - re-sort
		}
			
		this.updateFoxClocksState(true);
				
		FoxClocks_Logger.log("-FoxClocks::onAddCmd()");
	},
	
	// ====================================================================================
	onAddAsCmd : function()
	{		
		var location = this.getSelectedZonePickerLocation();
		if (!location)
			return;
		
		var watchlistItem = new FoxClocks_WatchlistItem(location);
		var retLocation = this.openZoneInfo(watchlistItem, "ADD_AS");			
			
		if (retLocation)
		{			
			watchlistItem.location = retLocation;
			
			var addedItems = new Array();
			addedItems.push(watchlistItem);
		
			this.addToWatchlist(addedItems);
			this.sortWatchlist(); // AFM - if we're sorting, re-sort
			this.updateFoxClocksState(true);
		}
	},
	
	// ====================================================================================
	onRemoveCmd : function()
	{
		FoxClocks_Logger.log("+FoxClocks::onRemoveCmd()");
		
		var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
		var promptText = document.getElementById("fc-remove-confirm-label").getAttribute("label");  
		         
		var confirmEnabled = FoxClocks_PrefManager.getPref("foxclocks.watchlist.remove.confirm.enabled");
		
		var result = confirmEnabled ? promptService.confirm(window, null, promptText) : true;
		
		if (result == false)
			return;
			
		var indexesToRemove = new Array();
											
		var rangeCount = this.watchlistTree.view.selection.getRangeCount();
		for (var i=0; i < rangeCount; i++)
		{
			var start = {};
			var end = {};
			this.watchlistTree.view.selection.getRangeAt(i, start, end);
			
			for (var c = start.value; c != -1 && c <= end.value; c++)
			{
				indexesToRemove.push(c);
			}
		}
			
		var watchlist = FoxClocks_WatchlistManager.getWatchlist();
		var treeChildrenRoot = document.getElementById("fc-watchlist-treechildren-root");

		// AFM - suppress select events on the watchlistTree until watchlistTree and the watchlist are
		// synchronised - we get called back in updateFoxClocksState() (the onselect handler)
		// immediately after we set the suppressed flag back to false
		//
		this.watchlistTree.view.selection.selectEventsSuppressed = true;
		for (var i=indexesToRemove.length - 1; i >= 0; i--)
		{
			var indexToRemove = indexesToRemove[i];
			var item = this.watchlistTree.view.getItemAtIndex(indexToRemove);
			item.parentNode.removeChild(item); 
			watchlist[indexToRemove] = null;
		}
		
		// AFM - sorry about this
		//
		var j = 0;
				
		while (j < watchlist.length)
		{	
			if (watchlist[j] == null)
				watchlist.splice(j, 1);
			else
				j++;
		}
		
		this.watchlistTree.view.selection.selectEventsSuppressed = false;
					
		// this.updateFoxClocksState(true); - not necessary see above comment, but need to notify, since
		// onselect handler will not do this
		//
		this.setWatchlistPref();
		
		FoxClocks_Logger.log("-FoxClocks::onRemoveCmd()");
	},
	
	// ====================================================================================
	onMoveUpCmd : function()
	{
		// AFM - we know only one item is selected and moving up is possible, saccess to the event is
		// disabled otherwise
		//
		var selectedItem = this.getSelectedWatchlistItem();
		var prevItem = selectedItem.previousSibling;
		this.swapWatchlistItems(prevItem, selectedItem);
		
		this.setWatchlistUnsorted();

		// AFM - reselect the selected item - the selection is lost moving up. This will trigger
		// this.updateFoxClocksState(false)
		//
		this.watchlistTree.view.selection.select(this.watchlistTree.view.getIndexOfItem(selectedItem));
		
		// AFM - need to set, since onselect handler will not do this
		//
		this.setWatchlistPref();
	},
	
	// ====================================================================================
	onMoveDownCmd : function()
	{	
		// AFM - we know only one item is selected and moving down is possible, access to the event is
		// disabled otherwise
		//
		var selectedItem = this.getSelectedWatchlistItem();
		var nextItem = selectedItem.nextSibling;
		this.swapWatchlistItems(selectedItem, nextItem);
		
		this.setWatchlistUnsorted();
		this.updateFoxClocksState(true);
	},
	
	// ====================================================================================
	moveInWatchlist : function(moveToIndex)
	{
		// AFM - move selected items in the watchlist to moveToIndex
		//
		FoxClocks_Logger.log("+FoxClocks::moveInWatchlist(): " + moveToIndex);
							
		var indexesToRemove = new Array();
		var itemsToAdd = new Array();
		var watchlist = FoxClocks_WatchlistManager.getWatchlist();
			
		if (moveToIndex == null)
			moveToIndex = watchlist.length;
		
		var numSelectedAboveMoveToIndex = 0;								
		var rangeCount = this.watchlistTree.view.selection.getRangeCount();
		for (var i=0; i < rangeCount; i++)
		{
			var start = {};
			var end = {};
			this.watchlistTree.view.selection.getRangeAt(i, start, end);
			
			for (var c = start.value; c != -1 && c <= end.value; c++)
			{
				indexesToRemove.push(c);
				itemsToAdd.push(watchlist[c]);
				
				if (c < moveToIndex)
					numSelectedAboveMoveToIndex++;
			}
		}
	
		for (var i=indexesToRemove.length - 1; i >= 0; i--)
		{
			var indexToRemove = indexesToRemove[i];
			watchlist[indexToRemove] = null;
		}
		
		var j = 0;			
		while (j < watchlist.length)
		{	
			if (watchlist[j] == null)
				watchlist.splice(j, 1);
			else
				j++;
		}
		
		// AFM - adjust moveToIndex by the number of selected items above this index
		//
		var adjMoveToIndex = moveToIndex - numSelectedAboveMoveToIndex;
		for (var k=0; k < itemsToAdd.length; k++)
		{
			watchlist.splice(adjMoveToIndex + k, 0, itemsToAdd[k]);				
		}

		this.addToWatchlist(null); // AFM - rebuild
		this.setWatchlistUnsorted();
		this.updateFoxClocksState(true);
		
		FoxClocks_Logger.log("-FoxClocks::moveInWatchlist()");
	},
	
	// ====================================================================================
	onWatchlistSelect : function() { this.updateFoxClocksState(false); },
	onOpenOptionsCmd : function() { FoxClocks_Utils.openChromeOptions(); },
	onOpenAboutCmd : function() { FoxClocks_Utils.openChromeAbout(); },
	onSortWatchlist : function(event) { this.sortWatchlist(event); this.updateFoxClocksState(true); },
	
	// ====================================================================================
	onOpenLocDetailsCmd : function()
	{
		var locationItem = this.getSelectedWatchlistItem();
		if (locationItem == null)
			return;
			
		var locationIndex = this.watchlistTree.view.getIndexOfItem(locationItem);
		var watchlistItem = FoxClocks_WatchlistManager.getItem(locationIndex);
		
		var retLocation = this.openZoneInfo(watchlistItem, "LOCATION");
		
		if (retLocation)
		{
			watchlistItem.location = retLocation;
				
			this.addToWatchlist(null);					
			this.sortWatchlist();			
			this.updateFoxClocksState(true);
		}
	},
	
	// ====================================================================================
	onOpenLocGoogleEarthCmd : function()
	{
		var locationItem = this.getSelectedWatchlistItem();
		
		if (locationItem == null)
			return;
			
		var locationIndex = this.watchlistTree.view.getIndexOfItem(locationItem);
		FoxClocks_Utils.openGoogleEarth(FoxClocks_WatchlistManager.getItem(locationIndex).location);
	},
	
	// ====================================================================================
	onToggleBarClocksCmd : function()
	{
		FoxClocks_Logger.log("+FoxClocks::onToggleBarClocksCmd()");
		
		var rangeCount = this.watchlistTree.view.selection.getRangeCount();
		for (var i=0; i < rangeCount; i++)
		{
			var start = {};
			var end = {};
			this.watchlistTree.view.selection.getRangeAt(i, start, end);
			
			for (var c = start.value; c != -1 && c <= end.value; c++)
			{
				var watchlistItem = FoxClocks_WatchlistManager.getItem(c);
				watchlistItem.showClock_statusbar = !watchlistItem.showClock_statusbar;
			}
		}
		
		this.updateFoxClocksState(true);
		
		FoxClocks_Logger.log("+FoxClocks::onToggleBarClocksCmd()");
	},
	
	// ====================================================================================
	onToggleBarFlagCmd : function()
	{
		FoxClocks_Logger.log("+FoxClocks::onToggleBarFlagCmd()");
		
		var rangeCount = this.watchlistTree.view.selection.getRangeCount();
		for (var i=0; i < rangeCount; i++)
		{
			var start = {};
			var end = {};
			this.watchlistTree.view.selection.getRangeAt(i, start, end);
			
			for (var c = start.value; c != -1 && c <= end.value; c++)
			{
				var watchlistItem = FoxClocks_WatchlistManager.getItem(c);
				watchlistItem.showClock_statusbarFlag = !watchlistItem.showClock_statusbarFlag;
			}
		}
		
		this.updateFoxClocksState(true);
		
		FoxClocks_Logger.log("+FoxClocks::onToggleBarFlagCmd()");
	},
	
	// ====================================================================================
	onTimer : function()
	{
		// FoxClocks_Logger.log("+FoxClocks::onTimer()");
		
		var nowDate = new Date();		
		for (var j=0; j < this.watchlistTree.view.rowCount; j++)
		{			
			var treeItem = this.watchlistTree.view.getItemAtIndex(j);
			
			// AFM - historical - keeping around
			//
			if (treeItem.getAttribute("id") == "fc-watchlist-noitems-item")
				continue;
			
			var location = FoxClocks_WatchlistManager.getItem(j).location;
			var timeText = this.timeFormatter.getTimeString(location, nowDate);
			
			var timeCell = treeItem.firstChild.childNodes[1];
			timeCell.setAttribute("label", timeText);
		}
		
		// FoxClocks_Logger.log("-FoxClocks::onTimer()");
	},
	
	// ====================================================================================
	// AFM - implementing nsIObserver
	//	
	observe : function(subject, topic, data)
	{
		if (topic == "foxclocks")
		{
			if (data == "engine:zone-data-update-complete")
			{
				if (FoxClocks_UpdateManager.getLastUpdateResult().result == "OK_NEW")
				{
					this.populateZonePicker();
					this.populateWatchlist();
				}
			}
			else if (data == "engine:zone-picker-changed")
			{
				this.populateZonePicker();
			}
			else if (data == "engine:watchlist-changed")
			{
				this.populateWatchlist();
			}
		}
		else if (topic == "nsPref:changed")
		{
			switch (data)
			{
				case "foxclocks.watchlist": this.addToWatchlist(null); break;
				case "foxclocks.format.foxclocks.standard": this.setTimeFormat(); break;
				case "foxclocks.format.foxclocks.custom": this.setTimeFormat(); break;
				case "foxclocks.clock.style": this.updateFoxClocksState(false); break;
				case "foxclocks.clock.foxclocks.clock.global.showflag": this.addToWatchlist(null); break; // AFM - rebuild watchlist with/without flags
			}
		}
		
		this.onTimer();
	},
	
	//
	// AFM - END, EVENTS
	
	// ====================================================================================
	populateWatchlist : function()
	{
		// AFM - completely regenerate watchlist, when driven by external notification
		// (eg importing settings), or on load
		// Sort state is not persisted as a preference, so to be safe we remove the sort marker
		//
		this.addToWatchlist(null);
		this.setWatchlistUnsorted();
		this.updateFoxClocksState(false);
	},
	
	// ====================================================================================
	populateZonePicker : function()
	{
		FoxClocks_Logger.log("+FoxClocks::populateZonePicker()");

		var zonePickerDoc = FoxClocks_ZoneManager.getZonePickerXmlDoc();
		if (zonePickerDoc !== null)
		{
			var treeChildrenRoot = document.getElementById("fc-zonepicker-treechildren-root");
			
			while (treeChildrenRoot.hasChildNodes())
				treeChildrenRoot.removeChild(treeChildrenRoot.firstChild);
		
			var zpRoot = zonePickerDoc.documentElement;
			for (var i=0; i < zpRoot.childNodes.length; i++)
			{
				this.populateZonePickerFromNode(treeChildrenRoot, zpRoot.childNodes[i]);
			}
		}		
		
		this.setZonePickerStates();
		
		FoxClocks_Logger.log("-FoxClocks::populateZonePicker()");
	},
		
	// ====================================================================================
	populateZonePickerFromNode : function(parentTreeChildren, zpNode)
	{
		var zonePickerLocMap = FoxClocks_ZoneManager.getZonePickerLocationMap();	
		var nowDate = new Date();
	
		// FoxClocks_Logger.log("FoxClocks::populateZonePickerFromNode(): " + zpNode.nodeName);		

		if (zpNode.nodeType != Node.ELEMENT_NODE)
			return;
			
		var zpNodeNameAtt = zpNode.getAttribute("name");
		
		if (zpNode.nodeName == "Leaf")
		{
			var leafId = zpNode.getAttribute("leaf_id");
		
			// AFM - this is a bad node, with no corresponding entry in zonePickerLocMap
			//	
			if (leafId == null)
				return; 
		
			var currLocation = zonePickerLocMap[leafId];
			var instantInfo = currLocation.zone.getInstantInfo(nowDate);
				
			var treeItem2 = document.createElement("treeitem");
			treeItem2.setAttribute("value", leafId); // AFM - to allow lookup of location from treeitem
			
			parentTreeChildren.appendChild(treeItem2);
			
			var cellCurrZoneLabel = instantInfo.offset + " (" + instantInfo.name + ")";

			if (instantInfo.is_dst)
				cellCurrZoneLabel += this.ZP_DST_INDICATOR_STRING;
			
			var treeCellRegion = document.createElement("treecell");
			treeCellRegion.setAttribute("label", zpNodeNameAtt);

			// var flagURL = currLocation.zone.getFlagUrl();	
			// if (FoxClocks_Utils.isUriAvailable(flagURL))
			//	treeCellRegion.setAttribute("src", flagURL);
			
			var treeCellCurrZone = document.createElement("treecell");
			treeCellCurrZone.setAttribute("label", cellCurrZoneLabel);
			
			var treeRow2 = document.createElement("treerow");					
			treeRow2.appendChild(treeCellRegion);
			treeRow2.appendChild(treeCellCurrZone);
			
			treeItem2.appendChild(treeRow2);
				
			// AFM - set search text
			//
			var leafNodeSearchText = treeCellRegion.getAttribute("label");
			var currentUpperTreeItem = treeItem2.parentNode.parentNode;
			
			while (currentUpperTreeItem.nodeName == "treeitem")
			{
				leafNodeSearchText = leafNodeSearchText + ' ' + currentUpperTreeItem.firstChild.firstChild.getAttribute("label");
				currentUpperTreeItem = currentUpperTreeItem.parentNode.parentNode;
			}
			
			treeCellRegion.setAttribute("fc_searchtext", leafNodeSearchText.toLowerCase());
			// treeCellRegion.setAttribute("label", treeCellRegion.getAttribute("fc_searchtext"));
		}
		else if (zpNode.nodeName == "Branch")
		{
			var childTreeChildren = document.createElement("treechildren");
			
			var treeItem1 = document.createElement("treeitem");
			treeItem1.setAttribute("container", "true");
			
			parentTreeChildren.appendChild(treeItem1);
			
			var treeCell1 = document.createElement("treecell");
			treeCell1.setAttribute("label", zpNodeNameAtt);
			
			var treeRow1 = document.createElement("treerow");
			
			treeRow1.appendChild(treeCell1);
			treeItem1.appendChild(treeRow1);
			treeItem1.appendChild(childTreeChildren);
			
			for (var j=0; j < zpNode.childNodes.length; j++)
			{
				this.populateZonePickerFromNode(childTreeChildren, zpNode.childNodes[j]);
			}
		}
	},

	// ====================================================================================
	onZonePickerSearchFocus : function()
	{
		this.searchBox.setAttribute("class", "fc-zonepicker-searchbox-active " +
			this.searchBox.getAttribute("class"));
		
		if (this.searchBox.value == this.searchBox.getAttribute("fc_init_value"))
			this.searchBox.value = "";
	},
	
	// ====================================================================================
	onZonePickerSearchBlur : function()
	{		
		if (this.searchBox.value == "")
		{
			this.searchBox.value = this.searchBox.getAttribute("fc_init_value");
			this.searchBox.setAttribute("class", "fc-zonepicker-searchbox-inactive");
		}
	},
	
	// ====================================================================================
	onZonePickerSearchInput : function()
	{
		// FoxClocks_Logger.log("+FoxClocks::onZonePickerSearchInput()");
		
		var inputSearchText = this.searchBox.value.replace(/(^\s+|\s+$)/g, '').toLowerCase();
		var zpTreeItems = document.getElementById("fc-zonepicker-treechildren-root").childNodes;
		var containerStack = new Array();
		var containerStacksToOpen = new Array();
		
		for (var i = 0; i < zpTreeItems.length; i++)
		{
			this.searchZonePickerNode(zpTreeItems[i], inputSearchText, containerStack, containerStacksToOpen);
		}
	
		if (containerStacksToOpen.length <= FoxClocks_Utils.FC_FOXCLOCKS_SEARCH_MAX_OPEN_NODES)
		{
			for (var k = 0; k < containerStacksToOpen.length; k++)
			{	
				var containerStack = containerStacksToOpen[k];
				for (var j = 0; j < containerStack.length; j++)
				{
					var containerItem = containerStack[j];
					var containerItemIndex = this.zonePicker.view.getIndexOfItem(containerItem);
					
					if (containerItemIndex != -1 && this.zonePicker.view.isContainer(containerItemIndex) &&
							this.zonePicker.view.isContainerOpen(containerItemIndex) == false)
					{
		         		this.zonePicker.view.toggleOpenState(containerItemIndex);
		         		// FoxClocks_Logger.log("FoxClocks::onZonePickerSearchInput(): opened <" + containerItem.firstChild.firstChild.getAttribute("label") + ">");
		         	}
				}
			}
		}
		else
		{
			FoxClocks_Logger.log("FoxClocks::onZonePickerSearchInput(): containerStacksToOpen.length is <" + containerStacksToOpen.length + ">");
		}
	
		var allNodesHidden = true;
		for (var l = 0; allNodesHidden == true && l < zpTreeItems.length; l++)
		{
			if (this.zonePicker.view.getIndexOfItem(zpTreeItems[l]) != -1)
				allNodesHidden = false;
		}
		
		if (allNodesHidden)
		{
			this.searchBox.setAttribute("class", "fc-zonepicker-searchbox-no-match");
		}
		else
		{
			this.searchBox.setAttribute("class", "fc-zonepicker-searchbox-active");
		}
			   	
		// FoxClocks_Logger.log("-FoxClocks::onZonePickerSearchInput(): inputSearchText <" + inputSearchText + ">");
	},
	
	// ====================================================================================
	searchZonePickerNode : function(treeItem, inputSearchText, containerStack, containerStacksToOpen)
	{
		// FoxClocks_Logger.log("+FoxClocks::searchZonePickerNode()");	
	
		var shouldShow = false;
		
		if (treeItem.getAttribute("container") == "true")
		{	
			containerStack.push(treeItem);
			
			var childTreeItems = treeItem.lastChild.childNodes;
			for (var i = 0; i < childTreeItems.length; i++)
			{
				if (this.searchZonePickerNode(childTreeItems[i], inputSearchText, containerStack, containerStacksToOpen) == true)
					shouldShow = true;
			}
		}
		else
		{	
			var cellSearchText = treeItem.firstChild.firstChild.getAttribute("fc_searchtext");
		
			if (inputSearchText == '' || cellSearchText.indexOf(inputSearchText) != -1)
			{ 	         	
				shouldShow = true;
				containerStacksToOpen.push(containerStack);
         		
				// if (inputSearchText.length > 3)
				//	FoxClocks_Logger.log("FoxClocks::searchZonePickerNode(): inputSearchText <" + inputSearchText + "> matches <" + cellSearchText + ">");
			}
		}
												
		if (shouldShow)
			treeItem.removeAttribute("hidden");
		else
			treeItem.setAttribute("hidden", "true");
		
		// FoxClocks_Logger.log("-FoxClocks::searchZonePickerNode()");
		
		return shouldShow;
	},
	
	// ====================================================================================
	addToWatchlist : function(addedItems, addAtIndex)
	{
		// AFM - addedItems == null => rebuild from FoxClocks_WatchlistManager.getWatchlist()
		// addAtIndex == null => add to end of list
		//
		FoxClocks_Logger.log("+FoxClocks::addToWatchlist()");
		var globaFoxClocksShowflagPref = FoxClocks_PrefManager.getPref("foxclocks.clock.foxclocks.clock.global.showflag");
		
		// AFM - to prevent modifying the tree triggering multiple select events; we get one
		// onselect event just after we unsuppress
		//
		this.watchlistTree.view.selection.selectEventsSuppressed = true;
	
		var watchlist = FoxClocks_WatchlistManager.getWatchlist();	
		var treeChildrenRoot = document.getElementById("fc-watchlist-treechildren-root");
		var itemsToAdd = addedItems;
		var rebuild = addedItems == null;
		
		if (rebuild)
		{
			while (treeChildrenRoot.hasChildNodes())
				treeChildrenRoot.removeChild(treeChildrenRoot.firstChild);
				
			itemsToAdd = watchlist;
		}
		
		var insertBeforeNode = addAtIndex != null ? treeChildrenRoot.childNodes[addAtIndex] : null;

		for (var i=0; itemsToAdd != "" && i < itemsToAdd.length; i++)
		{
			var currItemToAdd = itemsToAdd[i];
			
			var locationCell = document.createElement("treecell");
			locationCell.setAttribute("label", currItemToAdd.location.getName());													
									
			// AFM - per-clock not supported right now
			//
			if (globaFoxClocksShowflagPref != "fc-no-clocks")
			{
				var flagURL = currItemToAdd.getFlagUrl();	
				if (FoxClocks_Utils.isUriAvailable(flagURL))
					locationCell.setAttribute("src", flagURL);
			}

			var treeRow = document.createElement("treerow");
			treeRow.appendChild(locationCell);
			treeRow.appendChild(document.createElement("treecell"));
					
			var treeItem = document.createElement("treeitem");
			treeItem.setAttribute("value", currItemToAdd.location.zone.id);
			treeItem.appendChild(treeRow);

			if (insertBeforeNode != null)
			{
				treeChildrenRoot.insertBefore(treeItem, insertBeforeNode);
				
				if (!rebuild) // AFM - genuinely inefficient
					watchlist.splice(addAtIndex + i, 0, currItemToAdd);
			}
			else
			{
				treeChildrenRoot.appendChild(treeItem);	

				if (!rebuild)
					watchlist.push(currItemToAdd);
			}				
		}
		
		this.watchlistTree.view.selection.selectEventsSuppressed = false;
		FoxClocks_Logger.log("-FoxClocks::addToWatchlist()");
	},

	// ====================================================================================
	getSupportedFlavours : function()
	{
		// AFM - we could be observing drags from the Watchlist or ZonePicker, so we
		// can't rule out any flavours yet - this is done in onDragOver()
		//
		var flavours = new FlavourSet();
		flavours.appendFlavour("foxclocks/watchlist");
		flavours.appendFlavour("foxclocks/zonepicker");
		return flavours;
	},

	// ====================================================================================
	onDragStart: function(event, transferData, action)
	{
		FoxClocks_Logger.log("+FoxClocks::onDragStart()");

		var evtTargetId = event.target.getAttribute("id");		
		var flavour = "unknown";

		if (evtTargetId == "fc-watchlist-treechildren-root")
			flavour = "foxclocks/watchlist";
		else if (evtTargetId == "fc-zonepicker-treechildren-root")
			flavour = "foxclocks/zonepicker";
			
		transferData.data = new TransferData();
		transferData.data.addDataForFlavour(flavour, flavour); // AFM drop data is just the flavour again
	
		FoxClocks_Logger.log("-FoxClocks::onDragStart(): " + flavour);
	},
	
	// ====================================================================================
	onDragOver: function(event, flavour, session)
	{
		// FoxClocks_Logger.log("+FoxClocks::onDragOver(): " + event.target.getAttribute("id"));
		
		this.removeDropIndicator();
		this.watchlistDropAtIndex = null;
		
		session.canDrop = false; // AFM - unless we can show otherwise
		
		// AFM - Watchlist row we're currently over
		//
		var row = {}, col = {}, childElt = {};

		// AFM - handle weird Firefox 3.1 beta bug
		//
		if (this.watchlistTree.boxObject == null || this.watchlistTree.boxObject.getCellAt == null)
		{
			// FoxClocks_Logger.log("FoxClocks::onDragOver(): getCellAt() not available");
			session.canDrop = true;
			return;
		}
		
		this.watchlistTree.boxObject.getCellAt(event.clientX, event.clientY, row, col, childElt);
			    					
		var fromWatchlist = flavour.contentType == "foxclocks/watchlist";
		var toWatchlist = event.target.getAttribute("id") == "fc-watchlist-treechildren-root";
	
		// AFM - can't drag from ZonePicker to ZonePicker. Need to look at the selection count
		// if dragging from the ZonePicker, since a drag can start by dragging a ZP branch
		//
		if (!fromWatchlist && toWatchlist && this.zonePicker.view.selection.count > 0)
		{
			session.canDrop = true;
		}
		else if (fromWatchlist && toWatchlist && this.watchlistTree.view.selection.count > 0)
		{
			const disallowDropOnSelectedRow = true; // AFM - configuration
			if (disallowDropOnSelectedRow)
			{
				// AFM - disallow drop on any selected row
				//
				var mouseOverRowIsSelected = false;
				
				var rangeCount = this.watchlistTree.view.selection.getRangeCount();
				for (var i=0; i < rangeCount && mouseOverRowIsSelected == false; i++)
				{
					var start = {};
					var end = {};
					this.watchlistTree.view.selection.getRangeAt(i, start, end);
					
					for (var c = start.value; c != -1 && c <= end.value && mouseOverRowIsSelected == false; c++)
					{
					    if (row.value == c)
							mouseOverRowIsSelected = true;
					}
				}
		
				if (mouseOverRowIsSelected == false)
					session.canDrop = true;
			}
			else
			{
				session.canDrop = true;
			}			
		}
		else if (fromWatchlist && !toWatchlist && this.watchlistTree.view.selection.count > 0)
		{
			session.canDrop = true;
		}
		
		// AFM - we're dragging over Watchlist - figure out which rows to drop between
		//
		if (toWatchlist && session.canDrop)
		{			
			var locationItem = null;
			var locationItemDropBelow = null;
			var treeChildrenRoot = document.getElementById("fc-watchlist-treechildren-root");
			
			if (row.value == -1)
			{
				// AFM - we're not over any row, so we're dropping at the end
				//
				this.watchlistDropAtIndex = null;
				
				if (treeChildrenRoot.childNodes.length > 0)
				{
					locationItem = treeChildrenRoot.lastChild;
					locationItemDropBelow = true;
				}				
			}
			else
			{
				// AFM - get treecell of row we're currently over
				//
				var x_unused = {}, rowY = {}, width_unused = {}, rowHeight = {};
				this.watchlistTree.boxObject.getCoordsForCellItem(row.value, col.value, "text", x_unused, rowY, width_unused, rowHeight);
				
				locationItem = this.watchlistTree.view.getItemAtIndex(row.value);
			
				// AFM - y coord of mouse relative to the treechildren root
				//
				var relaClientY = event.clientY - treeChildrenRoot.boxObject.y;
	
				if (relaClientY >= rowY.value + rowHeight.value/2)
				{
					locationItemDropBelow = true;
					
					if (row.value < treeChildrenRoot.childNodes.length - 1)
						this.watchlistDropAtIndex = row.value + 1;
					else
						this.watchlistDropAtIndex = null;
				}
				else
				{
					locationItemDropBelow = false;
					this.watchlistDropAtIndex = row.value;
				}
			}
			
			if (locationItem != null && locationItemDropBelow != null)
			{
				// AFM - style the row to indicate where we would drop - try to use 'above' style if possible,
				// to avoid possible transition effects going from eg bottom half of one item to the top
				// half of the next. Should have read http://www.xulplanet.com/tutorials/xultu/treestyle.html
				// before I did this, but in fact dropBefore and dropAfter don't seem to work automatically...
				//
				if (locationItemDropBelow == true)
				{
					if (locationItem.nextSibling != null)
						locationItem.nextSibling.firstChild.setAttribute("properties", "foxclocks-drop-above");
					else
						locationItem.firstChild.setAttribute("properties", "foxclocks-drop-below");
				}
				else
				{
					locationItem.firstChild.setAttribute("properties", "foxclocks-drop-above");
				}
			}
		}
		
		// FoxClocks_Logger.log("-FoxClocks::onDragOver(): " +
		// this.watchlistTree.view.selection.count + " " +
		// flavour.contentType + " " + event.target.getAttribute("id"));	
	},
	
	// ====================================================================================
	onDragExit: function(event, session)
	{
		FoxClocks_Logger.log("+FoxClocks::onDragExit()");
		
		this.removeDropIndicator();
		
		FoxClocks_Logger.log("-FoxClocks::onDragExit()");	
	},
	
	// ====================================================================================
	onDrop: function(event, dropdata, session)
	{
		FoxClocks_Logger.log("+FoxClocks::onDrop()");

		this.removeDropIndicator(); // AFM - before we modify the Watchlist
		
		var fromWatchlist = dropdata.data == "foxclocks/watchlist";
		var toWatchlist = event.target.getAttribute("id") == "fc-watchlist-treechildren-root";
			
		if (!fromWatchlist && toWatchlist)
			this.onAddCmd(this.watchlistDropAtIndex, false);
		else if (fromWatchlist && !toWatchlist)
			this.onRemoveCmd();
		else if (fromWatchlist && toWatchlist)
			this.moveInWatchlist(this.watchlistDropAtIndex);

		this.watchlistDropAtIndex = null;
		
		FoxClocks_Logger.log("-FoxClocks::onDrop()");
	},

	
	// ====================================================================================
	removeDropIndicator: function()
	{
		// FoxClocks_Logger.log("+FoxClocks::removeDropIndicator()");
			
		// AFM - subtle, clever stuff - blow away all possible 'where we would drop' styles
		//
		if (this.watchlistDropAtIndex != null)
		{			
			var locationItem = this.watchlistTree.view.getItemAtIndex(this.watchlistDropAtIndex);
			locationItem.firstChild.removeAttribute("properties");
			
			if (locationItem.nextSibling != null)
				locationItem.nextSibling.firstChild.removeAttribute("properties");
		}
		else
		{
			var treeChildrenRoot = document.getElementById("fc-watchlist-treechildren-root");

			if (treeChildrenRoot.childNodes.length > 0)
				treeChildrenRoot.lastChild.firstChild.removeAttribute("properties");
		}
		
		// FoxClocks_Logger.log("-FoxClocks::removeDropIndicator()");	
	},
	
	// ====================================================================================
	sortWatchlist : function(event)
	{
		// AFM - return code currently meaningless
		//	
		FoxClocks_Logger.log("+FoxClocks::sortWatchlist()");

		var reSort = event == null;
		var sortResource = this.watchlistTree.getAttribute("sortResource");

		if (reSort && sortResource == "") // AFM - re-sorting, but nothing to sort by
			return false;
			
		var column = reSort ? document.getElementById(sortResource) : event.target;
		var colId = column.getAttribute("id");
		var colSortDir = column.getAttribute("sortDirection");
		
		if (reSort)
		{
			newSortDir = colSortDir;
		}
		else
		{
			// AFM - (no sort -> ascending -> descending)
			//
	
			if (colSortDir == "descending")
			{
				this.setWatchlistUnsorted();
				return false;
			}
	
			var newSortDir = colSortDir == "ascending" ? "descending" : "ascending";
		}
		
		var currDate = new Date();
									
		function columnSort(aa, bb)
		{	
			function nameSort(a, b) {
				var aLower = a.location.getName().toLowerCase();
				var bLower = b.location.getName().toLowerCase();
				
				return	aLower > bLower ? 1 : (aLower == bLower ? 0 : -1); }
		
			function timeSort(a, b) {
				return	a.location.zone.getInstantInfo(currDate).offset_mins -
						b.location.zone.getInstantInfo(currDate).offset_mins; }
					
			var a = newSortDir == "ascending" ? aa : bb;
			var b = newSortDir == "ascending" ? bb : aa;
			
			var primarySort = colId == "fc_watchlist_col_location" ? nameSort : timeSort;
			var secondarySort = colId == "fc_watchlist_col_location" ? timeSort : nameSort;
			
			var retVal = primarySort(a, b);
			if (retVal == 0)
				retVal = secondarySort(aa, bb); // always ascending
				
			return retVal;
		}	
		
		var cols = this.watchlistTree.getElementsByTagName("treecol");
		for (var i = 0; i < cols.length; i++)
		{
			cols[i].removeAttribute("sortDirection");
		}
		
		column.setAttribute("sortDirection", newSortDir);
		this.watchlistTree.setAttribute("sortResource", colId);
		
		FoxClocks_WatchlistManager.getWatchlist().sort(columnSort);
		this.addToWatchlist(null); // AFM - rebuild watchlistTree - no notifications

		FoxClocks_Logger.log("-FoxClocks::sortWatchlist(): " + newSortDir);
		
		return false;
	},		
	
	// ====================================================================================
	setWatchlistUnsorted : function()
	{
		var cols = this.watchlistTree.getElementsByTagName("treecol");
		for (var i = 0; i < cols.length; i++)
		{
			cols[i].removeAttribute("sortDirection");
		}
		
		this.watchlistTree.removeAttribute("sortResource");
	},
	
	// ====================================================================================
	setTimeFormat : function()
	{
		var standardFormat = FoxClocks_PrefManager.getPref("foxclocks.format.foxclocks.standard");
		var customFormat = FoxClocks_PrefManager.getPref("foxclocks.format.foxclocks.custom");
										
		this.timeFormatter.setTimeFormat(standardFormat == "" ? customFormat : standardFormat);
	},
	
	// ====================================================================================
	setWatchlistPref : function()
	{		
		// AFM - we don't want to respond to our own update
		//
		var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
		
		try {observerService.removeObserver(this, "foxclocks");} catch(ex){}
		
		FoxClocks_PrefManager.removePrefObserver("foxclocks.", this);
		FoxClocks_PrefManager.setPref("foxclocks.watchlist", FoxClocks_WatchlistManager.watchlistToXmlString());
		FoxClocks_PrefManager.addPrefObserver("foxclocks.", this);
		
		observerService.addObserver(this, "foxclocks", false);

	},
	
	// ====================================================================================
	getTreeChildrenId : function(index, array) { return "fc-zonepicker-treechildren:" + array.slice(0, index + 1).join(""); },
	
	// ====================================================================================
	getSelectedZonePickerLocation: function()
	{
		if (this.zonePicker.view.selection.count != 1)
			return null;
			
		var start = {};
		var end = {};
		this.zonePicker.view.selection.getRangeAt(0, start, end);

		var zoneItem = this.zonePicker.view.getItemAtIndex(start.value);
		
		if (zoneItem.getAttribute("container") == "true")
			return null;

		return FoxClocks_ZoneManager.getZonePickerLocationMap()[zoneItem.getAttribute("value")];
	},
	
	// ====================================================================================
	getSelectedWatchlistItem : function()
	{
		if (this.watchlistTree.view.selection.count == 1)
		{
			var start = {};
			var end = {};
			this.watchlistTree.view.selection.getRangeAt(0, start, end);

			return this.watchlistTree.view.getItemAtIndex(start.value);
		}
		
		return null;
	},
	
	// ====================================================================================
	openZoneInfo : function(watchlistItem, mode)
	{
		var retVals = {location: null};
		
		window.openDialog(	"chrome://foxclocks/content/zoneinfo.xul", "",
							"chrome,modal,centerscreen,resizable=yes",
							this.timeFormatter, mode, watchlistItem, retVals);
							
		return retVals.location;
	},

	// ====================================================================================
	setZonePickerStates : function()
	{
		// AFM - enable/disable things that depend on what's selected in the zone picker
		//
		
		FoxClocks_Logger.log("+FoxClocks::setZonePickerStates()");
		
		var location = this.getSelectedZonePickerLocation();
		
		document.getElementById('cmd_fc_add').setAttribute("disabled", (this.zonePicker.view.selection.getRangeCount() == 0 ? "true" : "false"));
		document.getElementById('cmd_fc_addas').setAttribute("disabled", (location == null ? "true" : "false"));
		document.getElementById('cmd_fc_zonedetails').setAttribute("disabled", (location == null ? "true" : "false"));
		document.getElementById('cmd_fc_zonegoogleearth').setAttribute("disabled", (location == null || location.getLatitude() == null || location.getLongitude() == null ? "true" : "false"));
		
		FoxClocks_Logger.log("-FoxClocks::setZonePickerStates()");
	},
	
	// ====================================================================================
	updateFoxClocksState : function(setWatchlistPref)
	{
		// AFM - not an intuitive name. GUI changes dependent on changes to the watchlist
		// (cf setZonePickerStates()
		//
		FoxClocks_Logger.log("+FoxClocks::updateFoxClocksState()");
				
		var selectedCount = this.watchlistTree.view.selection.count;
		var selectedItem = this.getSelectedWatchlistItem();
		var selectedIndex = selectedItem ? this.watchlistTree.view.getIndexOfItem(selectedItem) : null;
		var prevItem = selectedItem ? selectedItem.previousSibling : null;
		var nextItem = selectedItem ? selectedItem.nextSibling : null;
		var location = selectedItem ? FoxClocks_WatchlistManager.getItem(selectedIndex).location : null;
		
		document.getElementById('cmd_fc_moveup').setAttribute("disabled", (prevItem == null ? "true" : "false"));
		document.getElementById('cmd_fc_movedown').setAttribute("disabled", (nextItem == null ? "true" : "false"));
		document.getElementById('cmd_fc_remove').setAttribute("disabled", (selectedCount == 0 ? "true" : "false"));
		document.getElementById('cmd_fc_locdetails').setAttribute("disabled", (selectedItem == null ? "true" : "false"));
		document.getElementById('cmd_fc_toggle_bar_clocks').setAttribute("disabled", (selectedCount == 0 ? "true" : "false"));
		document.getElementById('cmd_fc_locgoogleearth').setAttribute("disabled", (location == null || location.getLatitude() == null || location.getLongitude() == null ? "true" : "false"));
		
		if (setWatchlistPref) // AFM - typically want to do this
			this.setWatchlistPref();
			
		this.onTimer();
		
		FoxClocks_Logger.log("-FoxClocks::updateFoxClocksState()");
	},		
	
	// ====================================================================================
	swapWatchlistItems : function(firstItem, secondItem)
	{
		var firstLocationIndex = this.watchlistTree.view.getIndexOfItem(firstItem);
		var secondLocationIndex = this.watchlistTree.view.getIndexOfItem(secondItem);

		FoxClocks_Logger.log("FoxClocks::swapWatchlistItems(): " + firstLocationIndex + " " + secondLocationIndex);
		
		var firstWatchlistItem = FoxClocks_WatchlistManager.getItem(firstLocationIndex);
		var secondWatchlistItem = FoxClocks_WatchlistManager.getItem(secondLocationIndex);
		
		FoxClocks_WatchlistManager.setItem(firstLocationIndex, secondWatchlistItem);
		FoxClocks_WatchlistManager.setItem(secondLocationIndex, firstWatchlistItem);
		
		// AFM - swap items in watchlistTree
		//
		document.getElementById("fc-watchlist-treechildren-root").insertBefore(secondItem, firstItem);
	}
}