"use strict";

/* Copyright (c) 2012 Andy McDonald. All rights reserved. */
/* Please refer to licence.txt for licensing terms. */

// ====================================================================================
Components.utils.import("resource://foxclocks/engine.js");

// ====================================================================================
FoxClocks_Engine.start();
// Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer)
//	.initWithCallback(function() { FoxClocks_Engine.start(); }, 5000, Components.interfaces.nsITimer.TYPE_ONE_SHOT);

// ====================================================================================
function FoxClocks_Overlay()
{			
	this.clockTimeFormatter = new FoxClocks_TimeFormatter();
	this.tooltipTimeFormatter = new FoxClocks_TimeFormatter();
	this.clockUpdateTimer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
	
	this.clockTooltip = null;
	this.clockTooltipUpdate = false;
	this.clockContainerType = null;
	this.hCustomizeToolbarWindow = null;
}
	
// ====================================================================================
FoxClocks_Overlay.prototype =
{
	// ====================================================================================
	onLoad : function()
	{
		this.clockTooltip = document.getElementById("foxclocksoverlay-clock-tooltip");
		this.setClocksFormat();
		this.setTooltipFormat();
		this.setMenuItemState();
		
		var self = this;
		FoxClocks_Engine.onStarted(function() { self.startup(); } );
	},
		
	// ====================================================================================
	startup : function()
	{
		FoxClocks_Logger.log("+FoxClocks_Overlay::startup()");

		if (FoxClocks_Utils.shouldApplyStatusbarFix)
		{
			for each (var stylesheet in document.styleSheets)
			{
				if (stylesheet.href == "chrome://foxclocks/skin/foxclocks.css")
					stylesheet.insertRule(FoxClocks_Utils.FC_STATUSBAR_FIX_CSS, stylesheet.cssRules.length);
			}
		}
		
		this.setClocksContainerType('startup');
		
		this.createClocks();
		this.updateView();
		
		this.clockUpdateTimer.init(this, FoxClocks_Utils.FC_CLOCK_UPDATE_INTERVAL, Components.interfaces.nsITimer.TYPE_REPEATING_PRECISE);

		FoxClocks_PrefManager.addPrefObserver("foxclocks.", this);
				
		var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
		observerService.addObserver(this, "domwindowopened", false);
		observerService.addObserver(this, "domwindowclosed", false);
		observerService.addObserver(this, "foxclocks", false);
	
		FoxClocks_Logger.log("-FoxClocks_Overlay::startup()");
	},
	
	// ====================================================================================	
	onUnload : function()
	{
		this.clockUpdateTimer.cancel();
		
		FoxClocks_PrefManager.removePrefObserver("foxclocks.", this);

		var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
		try {observerService.removeObserver(this, "domwindowopened");} catch(ex){}
		try {observerService.removeObserver(this, "domwindowclosed");} catch(ex){}
		try {observerService.removeObserver(this, "foxclocks");} catch(ex){}
	},
	
	// ====================================================================================
	onClockIconClick : function(event) { if (event.button == 0) FoxClocks_Utils.openChromeFoxClocks(); },
	onClockDblClick : function(event) { if (event.button == 0) FoxClocks_Utils.openChromeFoxClocks(); },
	onClockTooltipShowing : function(event) { this.clockTooltipUpdate = true; this.updateView(); },
	onClockTooltipHidden : function(event) { this.clockTooltipUpdate = false; },
	
	// ====================================================================================
	// AFM - implementing nsIObserver
	//
	observe : function(subject, topic, data)
	{	
		// FoxClocks_Logger.log("FoxClocks_Overlay::observe(): topic: <" + topic + ">, clock location <" +
		//	this.clockContainerType + ">");
		
		if (topic == "foxclocks")
		{
			if (data == "engine:watchlist-changed")
			{
				this.createClocks();
			}
			else if (data == "engine:zone-data-update-complete")
			{		
				if (FoxClocks_UpdateManager.getLastUpdateResult().result == "OK_NEW")
					this.createClocks();
			}

		}
		else if (topic == "nsPref:changed")
		{
			switch (data)
			{
				case "foxclocks.clock.containertype": this.setClocksContainerType('clock_container_type_changed'); this.createClocks(); break;
				case "foxclocks.clock.style": this.setClocksVisibility(); break;
				case "foxclocks.clock.position.relative": this.createClocks(); break;
				case "foxclocks.format.clock.standard": this.setClocksFormat(); break;
				case "foxclocks.format.clock.custom": this.setClocksFormat(); break;
				case "foxclocks.format.tooltip.standard": this.setTooltipFormat(); break;
				case "foxclocks.format.tooltip.custom": this.setTooltipFormat(); break;
				case "foxclocks.clock.bar.clock.global.showflag": this.createClocks(); break;
				case "foxclocks.clock.tooltip.clock.global.showflag": this.createClocks(); break;
				case "foxclocks.toolbar.menuitem.hidden": this.setMenuItemState(); break;
			}
		}
		else if (topic == "domwindowopened") // AFM - doesn't fire on Mac
		{
			var self = this;
			subject.addEventListener("load", function() {
				if (FoxClocks_Utils.FC_REGEXP_CUSTOMIZE_TOOLBAR.test(subject.location.href))
				{
					FoxClocks_Logger.log("FoxClocks_Overlay::observe(): customize toolbar window opened");
					self.hCustomizeToolbarWindow = subject;

					self.createClocks();
					self.updateView();
				}
			}, false);
			
			return;
		}
		else if (topic == "domwindowclosed") // AFM - doesn't fire on Mac
		{
			if (this.hCustomizeToolbarWindow != null)
			{
				FoxClocks_Logger.log("FoxClocks_Overlay::observe(): customize toolbar window closed");		
				this.hCustomizeToolbarWindow = null;

				if (this.clockContainerType == "fc-clock-containertype-statusbar")
				{
					var toolbarButton = document.getElementById("foxclocksoverlay-toolbarbutton-clocks");
					if (toolbarButton != null && toolbarButton.hidden == false)
					{
						FoxClocks_Logger.log("FoxClocks_Overlay::observe(): detected move from statusbar to toolbar");
						FoxClocks_PrefManager.setPref("foxclocks.clock.containertype", "fc-clock-containertype-toolbar");
					}
				}
				else
				{
					this.createClocks();
				}
			}
		}

		this.updateView();
	},
	
	// ====================================================================================
	updateView : function()
	{			
		var nowDate = new Date();
		
		for (var i=0; i < FoxClocks_WatchlistManager.getWatchlist().length; i++)
		{
			var watchlistItem = FoxClocks_WatchlistManager.getItem(i);
			var clock = document.getElementById('foxclocksoverlay-clock-id-' + i);
		
			if (clock == null || clock.hidden)
				continue;
			
			var sbarFormatter = watchlistItem.getStatusbarTimeFormatter() != null ?
				watchlistItem.getStatusbarTimeFormatter() : this.clockTimeFormatter;
				
			clock.setAttribute("label", sbarFormatter.getTimeString(watchlistItem.location, nowDate));
			
			var colour = watchlistItem.colour;
			
			if (watchlistItem.altColour_enabled)
			{
				var localDate = new Date(nowDate.getTime() + watchlistItem.location.zone.getInstantInfo(nowDate).offset_mins * 1000 * 60);
				var minsPastMidnight = localDate.getUTCHours() * 60 + localDate.getUTCMinutes();
				
				var startTime = watchlistItem.altColour_startTime;
				var endTime = watchlistItem.altColour_endTime;
				
				if (startTime < endTime)
				{
					if (minsPastMidnight >= startTime && minsPastMidnight < endTime)
						colour = watchlistItem.altColour;
				}
				else
				{
					if (!(minsPastMidnight >= endTime && minsPastMidnight < startTime))
						colour = watchlistItem.altColour;
				}
			}

			var clockPanelLabel = this.getClockPanelLabel(clock);
			if (clockPanelLabel != null)
			{
				if (colour != "")
					clockPanelLabel.setAttribute("style", "color:" + colour + " !important;");
				else
					clockPanelLabel.removeAttribute("style");
			}
		}
	
		// AFM - update tooltip if visible
		//
		var clockTooltipGrid = this.clockTooltip.firstChild;
		if (clockTooltipGrid && this.clockTooltipUpdate)
		{
			var gridRows = clockTooltipGrid.lastChild;
			for (var j=0; j < gridRows.childNodes.length; j++)
			{
				var gridRow = gridRows.childNodes[j];
				var locationIndex = gridRow.getAttribute("value");
				
				if (locationIndex && locationIndex != "fc-nolocs-row")
				{
					var currWatchlistItem = FoxClocks_WatchlistManager.getItem(locationIndex);
					
					var tooltipFormatter = currWatchlistItem.getTooltipTimeFormatter() != null ?
						currWatchlistItem.getTooltipTimeFormatter() : this.tooltipTimeFormatter;
						
					var timeText = tooltipFormatter.getTimeString(currWatchlistItem.location, nowDate);					
											
					var pos = timeText.indexOf("\t");
					var label1Value = null;
					var label2Value = null;
					
					if (pos != -1)
					{
						label1Value = timeText.substring(0, pos);
						label2Value = timeText.substring(pos + 1).replace("\t", "", "g");
					}
					else
					{
						label1Value = "";
						label2Value = timeText;
					}
					
					gridRow.childNodes[2].lastChild.setAttribute("value", label1Value);
					gridRow.childNodes[3].lastChild.setAttribute("value", label2Value);
				}
			}
		}
	},
	
	// ====================================================================================
	setClocksContainerType : function(context)
	{
		this.clockContainerType = FoxClocks_PrefManager.getPref("foxclocks.clock.containertype");
		
		if (this.clockContainerType == "fc-clock-containertype-statusbar")
		{
			if	(	context == 'clock_container_type_changed' ||
					(	context == 'startup' &&
						FoxClocks_PrefManager.getPref("extensions." + FoxClocks_Utils.FC_GUID_FOXCLOCKS + ".statusbar.forceshow") == true
					)
				)
			{
				var unhidden = false;
				
				var addonBar = document.getElementById("addon-bar");
				if (addonBar != null && addonBar.getAttribute("collapsed") == "true")
				{	
					addonBar.setAttribute("collapsed", "false");
					document.persist("addon-bar", "collapsed");
					unhidden = true;
				}
				
				var statusBar = document.getElementById("status-bar");	
				if (statusBar != null && statusBar.getAttribute("hidden") == "true")
				{
					statusBar.setAttribute("hidden", "false");
					document.persist("status-bar", "hidden");	
					unhidden = true;
				}
				
				if (context == 'startup' && unhidden == true)
					FoxClocks_PrefManager.setPref("extensions." + FoxClocks_Utils.FC_GUID_FOXCLOCKS + ".statusbar.forceshow", false);
			}
		}
	},
	
	// ====================================================================================
	setClocksFormat : function()
	{
		var standardFormat = FoxClocks_PrefManager.getPref("foxclocks.format.clock.standard");
		var customFormat = FoxClocks_PrefManager.getPref("foxclocks.format.clock.custom");
										
		this.clockTimeFormatter.setTimeFormat(standardFormat == "" ? customFormat : standardFormat);
	},
	
	// ====================================================================================
	setTooltipFormat : function()
	{
		var standardFormat = FoxClocks_PrefManager.getPref("foxclocks.format.tooltip.standard");
		var customFormat = FoxClocks_PrefManager.getPref("foxclocks.format.tooltip.custom");
										
		this.tooltipTimeFormatter.setTimeFormat(standardFormat == "" ? customFormat : standardFormat);
	},
	
	// ====================================================================================
	setMenuItemState : function()
	{
		var hideMenuItem = FoxClocks_PrefManager.getPref("foxclocks.toolbar.menuitem.hidden");
		
		var menuItem = document.getElementById("foxclocksoverlay-menuitem");
		if (menuItem != null)
			menuItem.collapsed = hideMenuItem;
		
		var appMenuMenuItem = document.getElementById("foxclocksoverlay-menuitem-appmenu");
		if (appMenuMenuItem != null)
			appMenuMenuItem.collapsed = hideMenuItem;
	},
	
	// ====================================================================================
	setClocksVisibility : function()
	{
		// FoxClocks_Logger.log("FoxClocks_Overlay::setClocksVisibility(): container type: <"
		//		+ this.clockContainerType + ">. Watchlist length: <" + FoxClocks_WatchlistManager.getWatchlist().length + ">");
		
		var clockStyle = FoxClocks_PrefManager.getPref("foxclocks.clock.style");
		
		var showToolbarButton = (this.clockContainerType == "fc-clock-containertype-toolbar" && (clockStyle == "fc-clock-style-icon" || this.hCustomizeToolbarWindow != null));
		var toolbarButton = document.getElementById("foxclocksoverlay-toolbarbutton-clocks");
		
		if (toolbarButton)
			toolbarButton.setAttribute("hidden", showToolbarButton ? "false" : "true");
		
		var showStatusbarIcon = (this.clockContainerType == "fc-clock-containertype-statusbar" && clockStyle == "fc-clock-style-icon");
		var statusbarPanelIcon = document.getElementById("foxclocksoverlay-statusbarpanel-icon");
		
		if (statusbarPanelIcon)
			statusbarPanelIcon.setAttribute("hidden", showStatusbarIcon ? "false" : "true");
		
		for (var i=0; i < FoxClocks_WatchlistManager.getWatchlist().length; i++)
		{
			var clock = document.getElementById('foxclocksoverlay-clock-id-' + i);
			if (clock != null)
			{
				var item = FoxClocks_WatchlistManager.getItem(i);
				clock.hidden = (showToolbarButton || showStatusbarIcon || item == null || item.showClock_statusbar == false);
			}
		}
		
		// AFM - due to the 'persistent tooltip' Firefox bug, for now, don't show the tooltip if
		// there are no locations in it, rather than showing '(no locations selected)'. Exception is
		// when we're in icon mode (only point of icon mode is to see the tooltip)
		// Note that this bug only applies to the statusbar, but for consistency this workaround
		// isn't statusbar-specific
		//
		var clockTooltipGrid = this.clockTooltip.firstChild;
		if (clockTooltipGrid)
		{
			var gridRows = clockTooltipGrid.lastChild;
			
			if (	gridRows.childNodes.length == 1 && 
					gridRows.childNodes[0].getAttribute("value") == "fc-nolocs-row" && 
					clockStyle != "fc-clock-style-icon")		
			{			
				this.clockTooltip.setAttribute("collapsed", "true");
			}
			else
			{
				this.clockTooltip.removeAttribute("collapsed", "true");
			}
		}
	},

	// ====================================================================================
	createClocks : function()
	{
		FoxClocks_Logger.log("+FoxClocks_Overlay::createClocks()");

		var globalClockShowflagPref = FoxClocks_PrefManager.getPref("foxclocks.clock.bar.clock.global.showflag");
		var globalTooltipShowflagPref = FoxClocks_PrefManager.getPref("foxclocks.clock.tooltip.clock.global.showflag");
	
		// AFM - subtle: blow away the clocks and clockTooltipGrid from all possible places
		//
		var allClockContainerIds = ["status-bar", "foxclocksoverlay-toolbaritem-box"];
		
		for (var x=0; x < allClockContainerIds.length; x++)
		{
			var currClockContainer = document.getElementById(allClockContainerIds[x]);
			if (currClockContainer == null)
				continue;
			
			var foxclocksChildren = currClockContainer.getElementsByClassName('foxclocks');		
			for (var k = foxclocksChildren.length - 1; k >= 0; k--)
			{
				currClockContainer.removeChild(foxclocksChildren[k]);
			}
		}

		var clockTooltipGrid = this.clockTooltip.firstChild;
		
		if (clockTooltipGrid != null)
			this.clockTooltip.removeChild(clockTooltipGrid);
		
		var clockContainer = this.getClockContainer();
		if (!clockContainer)
		{
			// AFM - container dragged off toolbar, for example
			//
			this.setClocksVisibility();
			return;
		}

		var clockType = this.getClockXULType();
		var clocksPosnRelative = FoxClocks_PrefManager.getPref("foxclocks.clock.position.relative");
			
		// AFM - create tool tip grid/columns/column/rows elements
		//
		clockTooltipGrid = document.createElement("grid");
		clockTooltipGrid.setAttribute("flex", "1");
		this.clockTooltip.appendChild(clockTooltipGrid);
	
		var gridColumns = document.createElement("columns");
		gridColumns.appendChild(document.createElement("column"));
		gridColumns.appendChild(document.createElement("column"));
		gridColumns.appendChild(document.createElement("column"));
		clockTooltipGrid.appendChild(gridColumns);
		
		var gridRows = document.createElement("rows");
		clockTooltipGrid.appendChild(gridRows);
		
		var previousClock = null;
		var clockId = 0;
		
		var startIndex = this.clockContainerType == "fc-clock-containertype-statusbar" ? -1 : 0;
		for (var i = startIndex; i < FoxClocks_WatchlistManager.getWatchlist().length; i++)
		{			
			var clock = null;
			
			if (i == -1)
			{
				// AFM - -1 => must be in statusbar
				//
				clock = document.createElement(clockType);
				clock.setAttribute("id", "foxclocksoverlay-statusbarpanel-icon");
				clock.setAttribute("class", "statusbarpanel-iconic foxclocks");
				clock.setAttribute("onclick", "foxClocksOverlay.onClockIconClick(event)");
			}
			else
			{
				var watchlistItem = FoxClocks_WatchlistManager.getItem(i);
				
				// AFM - don't use i, in case we've 'continued' above
				//
				var clockIdString = "foxclocksoverlay-clock-id-" + clockId++;
				
				clock = document.createElement(clockType);
				clock.setAttribute("id", clockIdString);
				clock.setAttribute("ondblclick", "foxClocksOverlay.onClockDblClick(event)");			
				
				if (this.clockContainerType == "fc-clock-containertype-toolbar")
				{				
					var toolbarItem = document.getElementById("foxclocksoverlay-toolbaritem-clocks");
					var toolbarItemParent = toolbarItem && toolbarItem.parentNode;
					var toolbarItemGrandParent = toolbarItemParent && toolbarItemParent.parentNode;
					
					if (toolbarItemParent && toolbarItemGrandParent) // AFM - should always exist
					{
						var parentId = toolbarItemParent.getAttribute("id");
						var grandParentId = toolbarItemGrandParent.getAttribute("id");
						
						if (	!parentId || !grandParentId ||
								(parentId != "toolbar-menubar" &&
								grandParentId != "toolbar-menubar"))
						{
							clock.setAttribute("labelclass", "toolbarbutton-text foxclocks-toolbar-clock-text");
						}
					}
				}
				
				var classString = "foxclocks foxclocks-clock ";
				
				if (watchlistItem.bold == true) classString += "foxclocks-bold ";
				if (watchlistItem.italic == true) classString += "foxclocks-italic ";
				if (watchlistItem.underline == true) classString += "foxclocks-underline ";
					
				if (	globalClockShowflagPref != "fc-no-clocks" &&
						(	watchlistItem.showClock_statusbarFlag == true ||
							globalClockShowflagPref == "fc-all-clocks")
					)
				{
					var flagUrl = watchlistItem.getFlagUrl();
					if (flagUrl != "")
					{
						clock.setAttribute("src", flagUrl);
											
						if (this.clockContainerType == "fc-clock-containertype-statusbar")
							classString += "statusbarpanel-iconic-text";
					}
				}
				
				clock.setAttribute("class", classString);

				// AFM - generate the tool tip grid row
				//
				if (watchlistItem.showClock_statusbarTooltip == true)
				{
					var gridRow = document.createElement("row");
					
					// AFM - store index of location corresponding to tooltip row
					// so we can get the location's time when updating the tooltip
					//
					gridRow.setAttribute("value", i);
					gridRow.setAttribute("align", "center");
					
					var imageElement = document.createElement("image");				
					
					if (	globalTooltipShowflagPref != "fc-no-clocks" &&
							(	watchlistItem.showClock_statusbarTooltipFlag == true ||
								globalTooltipShowflagPref == "fc-all-clocks")
						)
					{
						var flagUrl = watchlistItem.getFlagUrl();
						if (flagUrl != "")
						{
							imageElement.setAttribute("src", flagUrl);
						}
					}
					
					gridRow.appendChild(imageElement);
					
					var label = document.createElement("label");
					label.setAttribute("value", watchlistItem.location.getName());
				
					gridRow.appendChild(label);
					
					var hbox = document.createElement("hbox");
					hbox.setAttribute("align", "right");
					hbox.setAttribute("class", "location");
					hbox.appendChild(document.createElement("label"));
					gridRow.appendChild(hbox);
					
					hbox = document.createElement("hbox");
					hbox.setAttribute("align", "left");
					hbox.setAttribute("class", "time");
					hbox.appendChild(document.createElement("label"));
					gridRow.appendChild(hbox);
	
					gridRows.appendChild(gridRow);
				}
			}
			
			// AFM - attributes common to the 'icon' clock and normal clocks
			//
			clock.setAttribute("context", "foxclocksoverlay-clock-context");
			clock.setAttribute("tooltip", "foxclocksoverlay-clock-tooltip");
			
			if (previousClock)
			{
				if (previousClock.nextSibling != null)
					clockContainer.insertBefore(clock, previousClock.nextSibling);
				else
					clockContainer.appendChild(clock);
			}
			else
			{				
				// AFM - first clock going in to the clockContainer
				//
				if (this.clockContainerType == "fc-clock-containertype-statusbar")
				{
					var statusbarIndex = null;
					
					if (clocksPosnRelative == "fc-clock-position-left")
						statusbarIndex = 0;
					else if (clocksPosnRelative == "fc-clock-position-right")
						statusbarIndex = clockContainer.childNodes.length;
					else
						statusbarIndex = Number(clocksPosnRelative);

					if (clockContainer.childNodes.length == 0 || statusbarIndex > clockContainer.childNodes.length - 1)
						clockContainer.appendChild(clock);
					else
						clockContainer.insertBefore(clock, clockContainer.childNodes[statusbarIndex]);
				}
				else //  "fc-clock-containertype-toolbar" or anything else
				{
					clockContainer.appendChild(clock);
				}
			}
			
			previousClock = clock;
		}
		
		if (gridRows.childNodes.length == 0)
		{
			// AFM - generate the tool tip grid row when there are no locations to be displayed
			// in the tooltip
			//
			var gridRow = document.createElement("row");
			gridRow.setAttribute("value", "fc-nolocs-row");
			
			var label = document.createElement("label");
			label.setAttribute("value", document.getElementById("foxclocksoverlay-nolocs-label").getAttribute("label"));
			label.setAttribute("disabled", "true");
			
			gridRow.appendChild(label);
			// AFM - don't add second column. Naughty but seems to make it look nicer...
			// gridRow.appendChild(document.createElement("label"));

			gridRows.appendChild(gridRow);
		}
				
		this.setClocksVisibility();
			
		FoxClocks_Logger.log("-FoxClocks_Overlay::createClocks()");
	},
	
	// ====================================================================================
	getClockContainer : function() { return document.getElementById(this.clockContainerType == "fc-clock-containertype-statusbar" ? "status-bar" : "foxclocksoverlay-toolbaritem-box"); },
	getClockXULType : function() { return this.clockContainerType == "fc-clock-containertype-statusbar" ? "statusbarpanel" : "foxclocks-clockpanel"; },
	getClockPanelLabel : function(clock) { return this.clockContainerType == "fc-clock-containertype-statusbar" ? 
			document.getAnonymousElementByAttribute(clock, "class", "statusbarpanel-text") :
			document.getAnonymousElementByAttribute(clock, "anonid", "label"); },
	
	// ====================================================================================
	// AFM - legacy code to support CuteMenus: http://www.extensionsmirror.nl/index.php?showtopic=4360
	//
	onMenuItemCmd : function(event) { FoxClocks_Utils.openChromeFoxClocks(); }
}