/* Copyright (c) 2012 Andy McDonald. All rights reserved. */
/* Please refer to licence.txt for licensing terms. */

// ====================================================================================
var EXPORTED_SYMBOLS = ["FoxClocks_Logger"];

// ====================================================================================
(function() {
	"use strict";
	
	// ====================================================================================
	const CI = Components.interfaces, CC = Components.classes, CU = Components.utils;

	const _DEBUG = 0;
	const _INFO = 1;
	const _WARN = 2;
	const _ERROR = 3;
	const _FATAL = 4;
	const _LEVEL_STRINGS =  ["DEBUG", "INFO", "WARN", "ERROR", "FATAL"];

	var _consoleService = CC['@mozilla.org/consoleservice;1'].getService(CI.nsIConsoleService);
			
	// ====================================================================================
	function Logger()
	{
		this._outStream = null;
		this._level = _WARN;
	}

	// ====================================================================================
	Logger.prototype =
	{	
		// ====================================================================================
		init: function(level, enableTempLog)
		{
			if (level !== null)
			{
				if (level >= _DEBUG && level <= _FATAL)
					this._level = level;
				else
					this.error("Logger::init(): invalid log level " + level);
			}
	
			if (enableTempLog === true && this._outStream === null)
			{	
				try
				{
					var tmpFile = CC["@mozilla.org/file/directory_service;1"].getService(CI.nsIProperties).get("TmpD", CI.nsILocalFile);										 
					tmpFile.append("foxclocks.log");
					tmpFile.createUnique(CI.nsIFile.NORMAL_FILE_TYPE, parseInt("0664", 8));
				
					this._outStream = CC["@mozilla.org/network/file-output-stream;1"].createInstance(CI.nsIFileOutputStream);
					this._outStream.init(tmpFile, 0x02 | 0x08 | 0x20, parseInt("0664", 8), 0);
					
					this.info("Logger::init(): log file: " + tmpFile.path);
				}
				catch (ex)
				{
					this._outStream = null;
					this.error("Logger::init(): log file initialization failed: " + ex);
				}
			}
			else if (enableTempLog === false && this._outStream !== null)
			{
				this._outStream.close();
				this._outStream = null;
				
				this.info("Logger::init(): log file disabled");
			}
		},

		// ====================================================================================
		_log: function(message, level)
		{		
			if (level < this._level)
				return;
					
			if (level <= _INFO)
			{
				_consoleService.logStringMessage("FoxClocks" + " (" + _LEVEL_STRINGS[level] + "): " + message);
			}
			else
			{
				var flag = (level == _WARN ? CI.nsIScriptError.warningFlag : CI.nsIScriptError.errorFlag);
	
				var error = CC["@mozilla.org/scripterror;1"].createInstance(CI.nsIScriptError);
				error.init("FoxClocks" + (level === _FATAL ? ' (FATAL): ' : ': ') + message, null, null, null, null, flag, null);
				_consoleService.logMessage(error);
			}
	
			if (this._outStream !== null)
			{
				var messageString = "FoxClocks (" + _LEVEL_STRINGS[level] + "): " + message + "\n";
				this._outStream.write(messageString, messageString.length);
			}
		},
		
		// ====================================================================================
		log: function(message) { this._log(message, _DEBUG); },
		info: function(message) { this._log(message, _INFO); },
		warn: function(message) { this._log(message, _WARN); },
		error: function(message) { this._log(message, _ERROR); },
		fatal: function(message) { this._log(message, _FATAL); }
	};

	// ====================================================================================
	this.FoxClocks_Logger = new Logger();

}).call(this);