/* Copyright (c) 2012 Andy McDonald. All rights reserved. */
/* Please refer to licence.txt for licensing terms. */

// ====================================================================================
var EXPORTED_SYMBOLS = ["FoxClocks_TimeFormatter"];

(function() {
	"use strict";

	// ====================================================================================
	const CI = Components.interfaces, CC = Components.classes, CU = Components.utils;
	CU.import("resource://foxclocks/logger.js");
	CU.import("resource://foxclocks/utils.js");

	// ====================================================================================
	var _formatFunctions = {};
	var _formatsRegexp = /(?:)/;
	
	(function() { // AFM - static init

		var bundleHash = {};
		var bundleEnum =  CC["@mozilla.org/intl/stringbundle;1"].getService(CI.nsIStringBundleService).createBundle("chrome://foxclocks/locale/foxclocks.properties").getSimpleEnumeration();
		while (bundleEnum.hasMoreElements())
		{
			var bundlePropElt = bundleEnum.getNext().QueryInterface(CI.nsIPropertyElement);
			bundleHash[bundlePropElt.key] = bundlePropElt.value;
		}
				
		var amLower = bundleHash['misc.am.lower'];
		var pmLower = bundleHash['misc.pm.lower'];
		var amUpper = bundleHash['misc.am.upper'];
		var pmUpper = bundleHash['misc.pm.upper'];
		var monthNameLongArray = bundleHash['misc.list.month.name.long'].split(',');
		var monthNameShortArray = bundleHash['misc.list.month.name.short'].split(',');
		var dayNameLongArray = bundleHash['misc.list.day.name.long'].split(',');	
		var dayNameShortArray = bundleHash['misc.list.day.name.short'].split(',');			
		var dayOrdinalArray = bundleHash['misc.list.day.ordinal'].split(',');
		var defaultLocation = bundleHash['misc.default.location'];
		var defaultOffset = bundleHash['misc.default.offset'];
		var defaultZoneName = bundleHash['misc.default.zonename'];
		
		// May not exist
		//
		var dayRelativeArray = bundleHash['misc.list.day.relative'] ? bundleHash['misc.list.day.relative'].split(',') : ['?', '?', '?'];
		var defaultDayRelativeString = bundleHash['options.format.custom.day.relative.label'] ? '[' + bundleHash['options.format.custom.day.relative.label'] + ']' : '?';
		var defaultAlignmentPosition = bundleHash['misc.default.alignpos'] ? bundleHash['misc.default.alignpos'] : '?';
	
		var formatLocaleConfig = {
			'options.format.custom.secs.value': function(date) { return FoxClocks_Utils.asTwoDigit(date.getUTCSeconds()); },
			'options.format.custom.mins.value': function(date) { return FoxClocks_Utils.asTwoDigit(date.getUTCMinutes()); },
			'options.format.custom.hours.12.1.value': function(date) { return FoxClocks_Utils.as12hr(date.getUTCHours()); },
			'options.format.custom.hours.12.2.value': function(date) { return FoxClocks_Utils.asTwoDigit(FoxClocks_Utils.as12hr(date.getUTCHours())); },
			'options.format.custom.hours.24.1.value': function(date) { return date.getUTCHours(); },
			'options.format.custom.hours.24.2.value': function(date) { return FoxClocks_Utils.asTwoDigit(date.getUTCHours()); },
			'options.format.custom.am.lower.value': function(date) { return date.getUTCHours() < 12 ? amLower : pmLower; },
			'options.format.custom.am.upper.value': function(date) { return date.getUTCHours() < 12 ? amUpper : pmUpper; },
			'options.format.custom.year.4.value': function(date) { return date.getUTCFullYear(); },
			'options.format.custom.year.2.value': function(date) { return date.getUTCFullYear().toString().substr(2, 2); },
			'options.format.custom.month.num.1.value': function(date) { return date.getUTCMonth() + 1; },
			'options.format.custom.month.num.2.value': function(date) { return FoxClocks_Utils.asTwoDigit(date.getUTCMonth() + 1); },
			'options.format.custom.month.name.long.value': function(date) { return monthNameLongArray[date.getUTCMonth()]; },
			'options.format.custom.month.name.short.value': function(date) { return monthNameShortArray[date.getUTCMonth()]; },
			'options.format.custom.day.num.1.value': function(date) { return date.getUTCDate(); },
			'options.format.custom.day.num.2.value': function(date) { return FoxClocks_Utils.asTwoDigit(date.getUTCDate()); },
			'options.format.custom.day.name.long.value': function(date) { return dayNameLongArray[date.getUTCDay()]; },
			'options.format.custom.day.name.short.value': function(date) { return dayNameShortArray[date.getUTCDay()]; },
			'options.format.custom.day.ordinal.value':function(date) { return dayOrdinalArray[date.getUTCDate() - 1]; },
			'options.format.custom.day.ofyear.value': function(date) { return _getDayOfYear(date); },
			'options.format.custom.day.ofyear.3.value': function(date) { return FoxClocks_Utils.asThreeDigit(_getDayOfYear(date)); },
			'options.format.custom.day.relative.value': _formatDayRelative(dayRelativeArray, defaultDayRelativeString),
			'options.format.custom.day.relative.numeric.value': _formatDayRelative(['-1', '', '+1'], '[-1/+1]'),
			'options.format.custom.other.location.value': function(date, instantInfo, location) { if (location) return location.getName(); else return defaultLocation; }, 
			'options.format.custom.other.offset.value': function(date, instantInfo) { if (instantInfo) return instantInfo.offset; else return defaultOffset; },
			'options.format.custom.other.zone.value': function(date, instantInfo) { if (instantInfo) return instantInfo.name; else return defaultZoneName; },
			'options.format.custom.other.dstindic.value': function(date, instantInfo) { if (instantInfo) return (instantInfo.is_dst ? "*" : ""); else return '[*]'; },
			'options.format.custom.other.alignpos.value': function(date, instantInfo) { return instantInfo ? "\t" : defaultAlignmentPosition; },
			'options.format.custom.week.iso.1.value': function(date) { return _getISOWeekNumber(date); },
			'options.format.custom.week.iso.2.value': function(date) { return FoxClocks_Utils.asTwoDigit(_getISOWeekNumber(date)); },
			'options.format.custom.week.na.1.value': function(date) { return _getNAWeekNumber(date); },
			'options.format.custom.week.na.2.value': function(date) { return FoxClocks_Utils.asTwoDigit(_getNAWeekNumber(date)); },
		};
		
		var formatsRegexpString = "";	
		for (var placeholderKey in formatLocaleConfig)
		{
			if (formatLocaleConfig.hasOwnProperty(placeholderKey) === false)
				continue;
			
			var placeholder = bundleHash[placeholderKey];
			if (typeof(placeholder) === 'undefined')
			{
				FoxClocks_Logger.warn("TimeFormatter::init(): skipping untranslated placeholder: " + placeholderKey);
				continue;
			}
			
			_formatFunctions[placeholder] = formatLocaleConfig[placeholderKey];
	
			if (formatsRegexpString !== '')
				formatsRegexpString += '|';
				
			formatsRegexpString += placeholder;
		}
	
		_formatsRegexp = new RegExp(formatsRegexpString, 'g');
	})();
		
	// ====================================================================================
	function TimeFormatter(formatString)
	{
		this.formatArray = [];
		this.setTimeFormat(formatString);
	}
	
	// ====================================================================================
	TimeFormatter.prototype =
	{
		getTimeString: function(location, date) { return _getTimeString(location, date, this.formatArray); },
		setTimeFormat: function(formatString) { this.formatArray = _createFormatArray(formatString); }
	};

	// ====================================================================================
	TimeFormatter.getTimeStringFromFormat = function(location, date, formatString) { return _getTimeString(location, date, _createFormatArray(formatString)); }
	TimeFormatter.getUTCTimeStringFromFormat = function(date, formatString) { return _getTimeString(null, date, _createFormatArray(formatString)); }
	TimeFormatter.getLocalTimeStringFromFormat = function(date, formatString) { return _getTimeString(null, date, _createFormatArray(formatString), true); }
	
	// ====================================================================================
	var _getTimeString = function(location, date, formatArray, defaultToLocal)
	{
		if (formatArray.length === 0)
			return '';

		var instantInfo = null;
		var offset_mins = 0;
		
		if (location != null)
		{
			instantInfo = location.zone.getInstantInfo(date);
			
			if (instantInfo !== null)
				offset_mins = instantInfo.offset_mins;
		}
		else if (defaultToLocal === true)
		{
			offset_mins = -1 * date.getTimezoneOffset();
		}

		var adjDate = new Date(date.getTime() + offset_mins * 1000 * 60);
					
		var timeString = "";
		for (var i=0; i < formatArray.length; i++)
		{
			var format = formatArray[i];
			timeString += typeof(format) === 'function' ? format(adjDate, instantInfo, location) : format;
		}
		
		return timeString;
	}

	// ====================================================================================
	var _createFormatArray = function(formatString)
	{
		var formatArray = [];
		
		if (typeof(formatString) === 'string')
		{
			var nextMatchIndex = 0;
	
			formatString.replace(_formatsRegexp, function(str, p1, p2, offset, s) {
				if (nextMatchIndex < p1)
				{
					var unmatchedText = formatString.substr(nextMatchIndex, (p1 - nextMatchIndex))
					formatArray.push(unmatchedText);
				}
				formatArray.push(_formatFunctions[str]);
	
				nextMatchIndex = p1 + str.length;
			});
			
			// AFM - check for text after the last match
			//
			if (nextMatchIndex <= formatString.length - 1)
				formatArray.push(formatString.substr(nextMatchIndex));
		}
		
		return formatArray;
	}

	// ====================================================================================
	function _getDayOfYear(date)
	{
		const DAYS_IN_MONTH_NORMAL = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];
		const DAYS_IN_MONTH_LEAP = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];
	
		var month = date.getUTCMonth();
		var year = date.getUTCFullYear();
		
		var isLeapYear = year % 4 == 0 && (year % 100 != 0 || year % 400 != 0);
		var monthArray = isLeapYear ? DAYS_IN_MONTH_LEAP : DAYS_IN_MONTH_NORMAL;
	
		var dayOfYear = date.getUTCDate();
		for (var i=0; i < month; i++)
		{
			dayOfYear += monthArray[i];
		}
		
		return dayOfYear;
	}
	
	// ====================================================================================
	function _getISOWeekNumber(date)
	{
		// AFM - Note transform to Monday-based day of week
		//
		var thurSameWeek = new Date(date.getFullYear(), date.getMonth(), date.getDate() - ((date.getDay() + 6) % 7) + 3);
	
		var firstThurYear = new Date(thurSameWeek.getFullYear(), 0, 4); // Note we use thurSameWeek's year
		firstThurYear.setDate(firstThurYear.getDate() - ((firstThurYear.getDay() + 6) % 7) + 3);
		
		return Math.ceil((thurSameWeek - firstThurYear) / 604800000) + 1; 
	}
	
	// ====================================================================================
	function _getNAWeekNumber(date)
	{
		var firstDayOfYear = new Date(date.getFullYear(), 0, 1);
		return Math.ceil((((date - firstDayOfYear) / 86400000) + firstDayOfYear.getDay() + 1) / 7);
	}
	
	//====================================================================================
	function _formatDayRelative(dayRelativeArray, defaultString)
	{
		return function(date, location)
		{
			if (!location)
				return defaultString;
			
			var dayNumber = Math.floor(date.getTime()/86400000);
			var localDayNumber = Math.floor(new Date().getTime()/86400000);
			return dayNumber === localDayNumber ? '' : (dayNumber < localDayNumber ? dayRelativeArray[0] : dayRelativeArray[2]);
		};
	}
	
	// ====================================================================================
	this.FoxClocks_TimeFormatter = TimeFormatter;

}).call(this);