/* Copyright (c) 2012 Andy McDonald. All rights reserved. */
/* Please refer to licence.txt for licensing terms. */

// ====================================================================================
var EXPORTED_SYMBOLS = ["FoxClocks_Zone", "FoxClocks_Location", "FoxClocks_ZoneManager"];

// ====================================================================================
(function() {
	"use strict";

	// ====================================================================================
	const CI = Components.interfaces, CC = Components.classes, CU = Components.utils;
	CU.import("resource://foxclocks/logger.js");
	CU.import("resource://foxclocks/utils.js");
	CU.import("resource://foxclocks/prefmanager.js");
	
	// ====================================================================================
	function Zone(zoneId)
	{
		this.id = zoneId;
		this.country_code = null;
		this.alias_for = null;
		this.transitions = null;
		this.fixed = null;
		
		this.comments_lang = 'en'; // AFM - hard-code - never going to be anything else...
		this.comments = null;
	
		this.st_name = "";
		this.st_to_string = "";
		
		this.dl_name = "";
		this.dl_to_string = "";
		
		this.dl_start_gmt = null;
		this.dl_end_gmt = null;
	
		this.defaultLocation = null;
	}
	
	// ====================================================================================
	Zone.referenceDate = new Date();
	
	// ====================================================================================
	Zone.prototype = 
	{	
		// ====================================================================================
		set : function(zone)
		{
			this.country_code = typeof(zone.country_code) === 'string' ? zone.country_code : null;
			this.alias_for = typeof(zone.alias_for) === 'string' ? zone.alias_for : null;
			this.transitions = zone.transitions != null ? zone.transitions : null;	
			this.fixed = zone.fixed != null ? zone.fixed : null;	
			this.comments = typeof(zone.comments) !== 'undefined' && typeof(zone.comments[this.comments_lang]) === 'string' ? zone.comments[this.comments_lang] : null;
		
			this.st_name = "???";
			this.st_to_string = "???";
			
			this.dl_name = "";
			this.dl_to_string = "";
			
			this.dl_start_gmt = null;
			this.dl_end_gmt = null;
		
			if (this.fixed != null)
			{
				this.st_name = this.fixed.name;
				this.st_to_string = this.fixed.offset.replace(/^-/, FoxClocks_Utils.FC_MINUS_SYMBOL);
			}
			else
			{
				var dateEpoch = Zone.referenceDate.getTime();
				var transitions = this.transitions;
				var transitionsLength = transitions.length;

				var dstBeginTransition = null;
				var dstBeginTransitionIndex = -1;
		
				var stdBeginTransition = null;
				var stdBeginTransitionIndex = -1;
				
				var transitionIndex = null;
		
				for (var i = 0; transitionIndex === null && i < transitionsLength; i++)
				{
					// Transition immediately before the first 'future' transition
					// Note >= - very startof the transition is considered to be in the previous
					// transition, meaning transitionIndex can be -1
					//
					if (transitions[i].epoch >= dateEpoch)
						transitionIndex = i - 1;
				}
				
				// No future transitions - use most recent
				//
				if (transitionIndex === null)
					transitionIndex = transitionsLength - 1;
				
				if (transitionIndex !== -1)
				{
					var transition = transitions[transitionIndex];
					
					if (transition.is_dst)
					{
						dstBeginTransition = transition;
						dstBeginTransitionIndex = transitionIndex;
					}
					else
					{
						stdBeginTransition = transition;
						stdBeginTransitionIndex = transitionIndex;
					}
					
					if (transitions.length > transitionIndex + 1)
					{
						transitionIndex++;
						
						transition = transitions[transitionIndex];
						
						if (transition.is_dst)
						{
							dstBeginTransition = transition;
							dstBeginTransitionIndex = transitionIndex;
						}
						else
						{
							stdBeginTransition = transition;
							stdBeginTransitionIndex = transitionIndex;
						}
					}
					
					if (dstBeginTransition !== null)
					{
						this.dl_name = dstBeginTransition.name;
						this.dl_to_string = dstBeginTransition.offset.replace(/^-/, FoxClocks_Utils.FC_MINUS_SYMBOL);
			
						this.dl_start_gmt = new Date(dstBeginTransition.epoch);
						if (transitions.length > dstBeginTransitionIndex + 1)
						{
							var dstEndTransition = transitions[dstBeginTransitionIndex + 1];
						
							if (dstEndTransition.is_dst === false) // virtually always true (otherwise name is a lie
								this.dl_end_gmt = new Date(dstEndTransition.epoch);
						}
					}
			
					if (stdBeginTransition !== null)
					{
						this.st_name = stdBeginTransition.name;		
						this.st_to_string = stdBeginTransition.offset.replace(/^-/, FoxClocks_Utils.FC_MINUS_SYMBOL);
					}
				}
			}
			
			this.defaultLocation = new Location(this, null,
					typeof(zone.coords) !== 'undefined' ? zone.coords.lat_decimal : null, 
					typeof(zone.coords) !== 'undefined' ? zone.coords.long_decimal : null);
			
			return this;
		},
	
		// ====================================================================================
		getInstantInfo : function(date)
		{
			var retVal = { name: '', offset: '', offset_mins: '', is_dst: null };
			var dateEpoch = date.getTime();
			
			if (this.fixed != null)
			{
				retVal.name = this.fixed.name;
				retVal.offset = this.fixed.offset;
				retVal.offset_mins = this.fixed.offset_mins;
				retVal.is_dst = false;
			}
			else
			{		
				var transitionIndex = null;
				var transitions = this.transitions;
				var transitionsLength = transitions.length;
		
				for (var i = 0; transitionIndex === null && i < transitionsLength; i++)
				{
					// Transition immediately before the first 'future' transition
					// Note >= - see comment above
					if (transitions[i].epoch >= dateEpoch)
						transitionIndex = i - 1; 
				}
				
				// No future transitions - use most recent
				//
				if (transitionIndex === null)
					transitionIndex = transitionsLength - 1;
				
				if (transitionIndex === -1)
					return null;
				
				var transition = transitions[transitionIndex];
				
				retVal.name = transition.name;
				retVal.offset = transition.offset;
				retVal.offset_mins = transition.offset_mins;
				retVal.is_dst = transition.is_dst;
			}
			
			return retVal;
		},
		
		// ====================================================================================
		getFlagUrl : function()
		{			
			return this.country_code != null ? "chrome://foxclocks/skin/flags/" + this.country_code.toLowerCase() + ".png" : "";
		}
	}
	
	// ====================================================================================
	// Location
	// ====================================================================================
	function Location(zone, name, lat, lng)
	{
		this.zone = zone;
		
		this._name = name;
		
		this._latitude = null;
		this._lat_degs = 0;
		this._lat_mins = 0;
		this._lat_secs = 0;
		
		this._longitude = null;
		this._long_degs = 0;
		this._long_mins = 0;
		this._long_secs = 0;
			
		this.setLatitude(lat);
		this.setLongitude(lng);
	}
	
	// ====================================================================================
	// TODO: ugh
	//
	var hFoxClocksBundle = CC["@mozilla.org/intl/stringbundle;1"].getService(CI.nsIStringBundleService).createBundle("chrome://foxclocks/locale/foxclocks.properties");
	
	Location.northAsLocaleString = hFoxClocksBundle.GetStringFromName("misc.north");
	Location.southAsLocaleString = hFoxClocksBundle.GetStringFromName("misc.south");
	Location.eastAsLocaleString = hFoxClocksBundle.GetStringFromName("misc.east");
	Location.westAsLocaleString = hFoxClocksBundle.GetStringFromName("misc.west");
	
	// ====================================================================================
	Location.prototype =
	{
		// ====================================================================================
		getLatitude : function() { return this._latitude; },
		getLongitude : function() { return this._longitude; },
		setName : function(name) { this._name = name; }, 
		getName : function() { return this._name !== null ? this._name : this.zone.defaultLocation._name; },
		
		// ====================================================================================
		setLatitude : function(lat)
		{
			// AFM - decimal degrees - 5dp - covers seconds
			//
			// AFM - weak typing: emptyString == 0 is true, and zero == "" is true. lat, lng
			// are sometimes strings, sometimes numbers
			//
			if (lat != null && ((typeof(lat) == "string" && lat != "") || typeof(lat) == "number"))
			{
				this._latitude = Math.round(lat*100000)/100000;
				
				var absLat = Math.abs(this._latitude);
				
				this._lat_degs = Math.floor(absLat);
				this._lat_mins = Math.floor(60 * (absLat - this._lat_degs));
				this._lat_secs = Math.round(3600 * (absLat - this._lat_degs - this._lat_mins/60));
				
				// AFM - may round up to 60
				//
				if (this._lat_secs == 60)
				{
					this._lat_mins++;
					this._lat_secs = 0;
				}	
			}
			else
			{
				this._latitude = null;
				this._lat_degs = 0;
				this._lat_mins = 0;
				this._lat_secs = 0;
			}
		},
	
		// ====================================================================================
		setLongitude : function(lng)
		{
			if (lng != null && ((typeof(lng) == "string" && lng != "") || typeof(lng) == "number"))
			{
				this._longitude = Math.round(lng*100000)/100000;
	
				var absLong = Math.abs(this._longitude);
				
				this._long_degs = Math.floor(absLong);
				this._long_mins = Math.floor(60 * (absLong - this._long_degs));
				this._long_secs = Math.round(3600 * (absLong - this._long_degs - this._long_mins/60));
				
				// AFM - may round up to 60
				//
				if (this._long_secs == 60)
				{
					this._long_mins++;
					this._long_secs = 0;
				}	
			}
			else
			{
				this._longitude = null;
				this._long_degs = 0;
				this._long_mins = 0;
				this._long_secs = 0;
			}
		},
		
		// ====================================================================================
		latitudeAsLocaleString : function()
		{
			// AFM - convert e.g. 52.367 to 52{deg. symbol}22' North
			//
			var latitudeAsString = "";
			if (this._latitude != null)
			{
				latitudeAsString = this.latitudeDegrees() + FoxClocks_Utils.FC_DEGREE_SYMBOL + FoxClocks_Utils.asTwoDigit(this.latitudeMins()) + "'";
				latitudeAsString += " " + (this.latitudeIsNorth() ? Location.northAsLocaleString : Location.southAsLocaleString);
			}
			
			return latitudeAsString;
		},
	
		// ====================================================================================
		longitudeAsLocaleString : function()
		{
			// AFM - convert e.g. -52.367 to 52{deg. symbol}22' West
			//
			var longitudeAsString = "";
			if (this._longitude != null)
			{				
				longitudeAsString = this.longitudeDegrees() + FoxClocks_Utils.FC_DEGREE_SYMBOL + FoxClocks_Utils.asTwoDigit(this.longitudeMins()) + "'";
				longitudeAsString += " " + (this.longitudeIsEast() ? Location.eastAsLocaleString : Location.westAsLocaleString);
			}
			
			return longitudeAsString;
		},
		
		// ====================================================================================
		latitudeIsNorth : function() { return this._latitude != null && this._latitude >= 0; },
		latitudeDegrees : function() { return this._lat_degs; },
		latitudeMins : function() { return this._lat_mins; },
		latitudeSecs : function() { return this._lat_secs; },
		longitudeIsEast : function() { return this._longitude != null && this._longitude >= 0; },
		longitudeDegrees : function() { return this._long_degs; },
		longitudeMins : function() { return this._long_mins; },
		longitudeSecs : function() { return this._long_secs; }
	}

	// ====================================================================================
	// ZoneManager
	// ====================================================================================
	function ZoneManager()
	{
		this.dataSource = {id: null, name: null, date: null, version: null};
		this.zoneData = null;
		
		this.zonePickerLocationMap = [];
		this.zonePickerXmlDoc = null;

		this.updateFile = 'resource://foxclocks-data/zones_update.json';
		this.dataFile = 'resource://foxclocks-data/zones.json';
	}
	
	// ====================================================================================
	ZoneManager.prototype =
	{	
		// ====================================================================================
		loadZoneData: function(callback)
		{
			var self = this;
			
			var loadUrl = function(url, loadCallback) {

				try
				{
					var req = CC["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(CI.nsIXMLHttpRequest);
					req.open('GET', url + '?time=' + new Date().getTime(), true);
					req.responseType = 'json';
					req.addEventListener('loadend', function() {
						
						var data = req.response;
						if (data === null || data.hasOwnProperty('zones') === false)
						{
							loadCallback(false, 'Bad response');
							return;
						}
						
						if (self.zoneData === null)
							self.zoneData = {};
						
						for (var zone_id in data.zones)
						{
							if (data.zones.hasOwnProperty(zone_id))
							{
								var zone = self.zoneData[zone_id];
								
								if (typeof(zone) === 'undefined')
								{
									zone = new Zone(zone_id);
									self.zoneData[zone_id] = zone;
								}
								
								zone.set(data.zones[zone_id]);
							}
						}
						
						self.dataSource = {id: data.source.id, name: data.source.name, date: data.source.date, version: data.source.version};
						
						FoxClocks_Logger.log("ZoneManager::loadZoneData(): loaded zone data from " + url);
						loadCallback(true);
						
					}, false);
					
					req.send(null);
				}
				catch(ex)
				{
					loadCallback(false, ex);
				}
			};

			loadUrl(self.updateFile, function(success, ex) {
				
				if (success === true)
				{
					callback(success);
				}
				else
				{
					FoxClocks_Logger.log("ZoneManager::loadZoneData(): could not load updated zone data from " + self.updateFile);

					loadUrl(self.dataFile, function(success, ex) {
					
						if (success === false)
							FoxClocks_Logger.error("ZoneManager::loadZoneData(): could not load built-in zone data from " + self.dataFile + ": " + ex);
						
						callback(success);
					});
				}
			});
		},
		
		// ====================================================================================
		loadZonePicker: function(callback, force)
		{
			if (this.zonePickerXmlDoc !== null && force !== true)
			{
				callback(true);
				return;
			}
	
			if (this.zoneData === null)
			{
				FoxClocks_Logger.error("ZoneManager::loadZonePicker(): no zone data");
				callback(false);
				return;
			}
			
			var self = this;
			this._loadZonePickerUrl(function(success) {
				
				if (success === false)
				{
					callback(success);
					return;
				}

				self.zonePickerLocationMap = [];

				var goodLeafNodeId = 0;
				var leafNodes = self.zonePickerXmlDoc.documentElement.getElementsByTagName("Leaf");
				
				for (var i=0; i < leafNodes.length; i++)
				{
					var leafNode = leafNodes.item(i);
					leafNode.QueryInterface(CI.nsIDOMElement);

					var name = leafNode.getAttribute("name");
					var zoneId = leafNode.getAttribute("zone_id");
					var zone = zoneId ? self.zoneData[zoneId] : null;
					
					if (name == null || zoneId == null || zone == null)
					{
						FoxClocks_Logger.error("ZoneManager::loadZonePicker(): cannot generate item with zone id <" + zoneId + "> - skipping");
						continue;
					}
					
					leafNode.setAttribute("leaf_id", goodLeafNodeId++);
					var defaultLocation = zone.defaultLocation;
					
					// AFM - default location's name was null on creation in loadZoneData(); this is the earliest
					// we can set it, in a locale-independent way
					//
					if (defaultLocation.getName() === null)
						defaultLocation.setName(name);
					
					var latitude = null;
					var longitude = null;
					
					var coords = leafNode.getAttribute("coords");
					if (coords !== null)
					{
						var coordsMatch = coords.match(/^([+-]?[0-9|\.]+), *([+-]?[0-9|\.]+)$/); // leave heavy lifting to parseFloat()
			
						if (coordsMatch !== null)
						{
							latitude = parseFloat(coordsMatch[1]);
							longitude = parseFloat(coordsMatch[2]);
						}
						else if (coords !== 'none' && coords !== 'na')
						{
							latitude = defaultLocation.getLatitude();
							longitude = defaultLocation.getLongitude();
						}
					}
					else
					{			
						// AFM - legacy support - old-style zonepicker.xml
						//
						var coordsNodes = leafNode.getElementsByTagName("Coordinates");
						if (coordsNodes.length == 1)
						{
							var coordsNode = coordsNodes.item(0);
							coordsNode.QueryInterface(CI.nsIDOMElement);
							
							latitude = coordsNode.getAttribute("latitude");
							longitude = coordsNode.getAttribute("longitude");
						}
						else
						{
							latitude = defaultLocation.getLatitude();
							longitude = defaultLocation.getLongitude();
						}
					}
				
					self.zonePickerLocationMap.push(new Location(zone, name, latitude, longitude));
				}
				
				callback(success);
			});

			FoxClocks_Logger.log("-ZoneManager::loadZonePicker()");
		},

		// ====================================================================================
		_loadZonePickerUrl: function(callback)
		{
			var self = this;
			
			var loadUrl = function(url, loadCallback) {

				try
				{
					var req = CC["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(CI.nsIXMLHttpRequest);
					req.open('GET', url + '?time=' + new Date().getTime(), true);
					req.responseType = 'xml';
					req.addEventListener('loadend', function() {
						
						if (req.responseXML === null)
							throw "Bad response";
						
						if (req.responseXML.documentElement.nodeName === "parsererror")
							throw req.responseXML.documentElement.firstChild.nodeValue;
				
						self.zonePickerXmlDoc = req.responseXML;
					
						FoxClocks_Logger.log("ZoneManager::_loadZonePickerUrl(): loaded zone picker from " + url);
						loadCallback(true);
						
					}, false);
					
					req.send(null);
				}
				catch(ex)
				{
					FoxClocks_Logger.log("ZoneManager::_loadZonePickerUrl(): could not load zone picker from " + url + ": " + ex);
					loadCallback(false);
				}
			};

			var zonePickerUrl = FoxClocks_PrefManager.getPref("foxclocks.zonepicker.dataurl");
			
			if (zonePickerUrl === "fc-zonepicker-dataurl-builtin")
				zonePickerUrl = FoxClocks_Utils.FC_URL_CHROME_ZONEPICKER_BUILTIN;
			
			loadUrl(zonePickerUrl, function(success) {
				
				if (success === true || zonePickerUrl === FoxClocks_Utils.FC_URL_CHROME_ZONEPICKER_BUILTIN)
				{
					callback(success);
				}
				else
				{
					loadUrl(FoxClocks_Utils.FoxClocks_Utils.FC_URL_CHROME_ZONEPICKER_BUILTIN, function(success) {
						callback(success);
					});
				}
			});
		},
		
		// ====================================================================================
		getZoneData: function() { return this.zoneData; },
		getZonePickerLocationMap: function() { return this.zonePickerLocationMap; },
		getZonePickerXmlDoc: function() { return this.zonePickerXmlDoc; },
	};

	// ====================================================================================
	this.FoxClocks_Zone = Zone;
	this.FoxClocks_Location = Location;
	this.FoxClocks_ZoneManager = new ZoneManager();

}).call(this);